import numpy as np
import pandas as pd

from rdt import get_demo


def test_get_demo():
    demo = get_demo()

    assert list(demo.columns) == [
        'last_login', 'email_optin', 'credit_card', 'age', 'dollars_spent'
    ]
    assert len(demo) == 5
    assert list(demo.isna().sum(axis=0)) == [1, 1, 1, 0, 1]


def test_get_demo_many_rows():
    demo = get_demo(10)

    login_dates = pd.Series([
        '2021-06-26', '2021-02-10', 'NaT', '2020-09-26', '2020-12-22', '2019-11-27',
        '2002-05-10', '2014-10-04', '2014-03-19', '2015-09-13'
    ], dtype='datetime64[ns]')
    email_optin = [False, False, False, True, np.nan, np.nan, False, True, False, False]
    credit_card = [
        'VISA', 'VISA', 'AMEX', np.nan, 'DISCOVER', 'AMEX', 'AMEX', 'DISCOVER', 'DISCOVER', 'VISA'
    ]
    age = [29, 18, 21, 45, 32, 50, 93, 75, 39, 66]
    dollars_spent = [99.99, np.nan, 2.50, 25.00, 19.99, 52.48, 39.99, 4.67, np.nan, 23.28]

    expected = pd.DataFrame({
        'last_login': login_dates,
        'email_optin': email_optin,
        'credit_card': credit_card,
        'age': age,
        'dollars_spent': dollars_spent
    })

    pd.testing.assert_frame_equal(demo, expected)



"""RDT code style module."""

import importlib
import inspect
import json
import re
from pathlib import Path
from types import FunctionType

import pytest

from rdt.transformers import TRANSFORMERS, get_transformer_class
from rdt.transformers.base import BaseTransformer


def validate_transformer_subclass(transformer):
    """Test whether or not the ``Transformer`` is a subclass of ``BaseTransformer``."""
    fail_message = 'Transformer must be a subclass of ``BaseTransformer``.'
    assert issubclass(transformer, BaseTransformer), fail_message


def validate_transformer_module(transformer):
    """Test whether or not the ``Transformer`` is inside the right module."""
    transformer_file = Path(inspect.getfile(transformer))
    transformer_folder = transformer_file.parent
    is_valid = False

    if transformer_folder.match('transformers'):
        is_valid = True
    elif transformer_folder.parent.match('transformers'):
        is_valid = True
    elif transformer_folder.parent.match('addons'):
        is_valid = True
    elif transformer_folder.parent.parent.match('addons'):
        is_valid = True

    assert is_valid, 'The transformer module is not placed inside a valid path.'


def _validate_config_json(config_path, transformer_name):
    with open(config_path, 'r', encoding='utf-8') as config_file:
        try:
            config_dict = json.load(config_file)
        except json.JSONDecodeError:
            config_dict = None

    config_error_msg = f'{config_path} does not have valid json format.'
    assert config_dict is not None, config_error_msg

    transformers_not_found = 'The key ``transformers``  was not found in the config.json file.'
    assert 'transformers' in config_dict, transformers_not_found
    assert 'name' in config_dict, 'The key ``name`` was not found in the config.json file.'

    transformers = config_dict.get('transformers', [])
    for transformer in transformers:
        module, name = transformer.rsplit('.', 1)
        if name == transformer_name:
            imported_transformer = getattr(importlib.import_module(module), name, None)
            assert imported_transformer is not None, f'Could not import {name} from {module}'


def validate_transformer_addon(transformer):
    """Assert if a ``Transformer`` is a valid ``addon``."""
    transformer_file = Path(inspect.getfile(transformer))
    transformer_folder = transformer_file.parent
    is_addon = transformer_folder.parent.match('addons')

    init_file_exist, module_py, config_json_exist = False, False, False

    if is_addon:
        documents = list(transformer_folder.iterdir())
        for document in documents:
            if document.match('__init__.py'):
                init_file_exist = True
            elif re.match(r'^[a-z]+.*.py$', document.name):
                module_py = True
            elif document.match('config.json'):
                config_json_exist = True
                _validate_config_json(document, transformer.__name__)

        assert init_file_exist, 'Missing __init__.py file within the addon folder.'
        assert config_json_exist, 'Missing the config.json file within the addon folder.'
        assert module_py, 'Missing addon module that contains the code.'


def validate_transformer_importable_from_parent_module(transformer):
    """Validate wheter the transformer can be imported from the parent module."""
    name = transformer.__name__
    module = getattr(transformer, '__module__', '')
    module = module.rsplit('.', 1)[0]
    imported_transformer = getattr(importlib.import_module(module), name, None)
    assert imported_transformer is not None, f'Could not import {name} from {module}'


def get_test_location(transformer):
    """Return the expected unit test location of a transformer."""
    transformer_file = Path(inspect.getfile(transformer))
    transformer_folder = transformer_file.parent
    rdt_unit_test_path = Path(__file__).parent / 'unit'

    test_location = None
    if transformer_folder.match('transformers'):
        test_location = rdt_unit_test_path / 'transformers' / f'test_{transformer_file.name}'

    elif transformer_folder.parent.match('transformers'):
        test_location = rdt_unit_test_path / 'transformers' / transformer_folder.name
        test_location = test_location / f'test_{transformer_file.name}'

    elif transformer_folder.parent.match('addons'):
        test_location = rdt_unit_test_path / 'transformers' / 'addons' / transformer_folder.name
        test_location = test_location / f'test_{transformer_file.name}'

    elif transformer_folder.parent.parent.match('addons'):
        test_location = rdt_unit_test_path / 'transformers' / 'addons'
        test_location = test_location / transformer_folder.parent.name / transformer_folder.name
        test_location = test_location / f'test_{transformer_file.name}'

    return test_location


def validate_test_location(transformer):
    """Validate if the test file exists in the expected location."""
    test_location = get_test_location(transformer)
    if test_location is None:
        return False, 'The expected test location was not found.'

    assert test_location.exists(), 'The expected test location does not exist.'


def _load_module_from_path(path):
    """Return the module from a given ``PosixPath``."""
    assert path.exists(), 'The expected test module was not found.'
    module_path = path.parent
    module_name = path.name.split('.')[0]
    if module_path.name == 'transformers':
        module_path = f'rdt.transformers.{module_name}'
    elif module_path.parent.name == 'transformers':
        module_path = f'rdt.transformers.{module_path.parent.name}.{module_name}'
    elif module_path.parent.name == 'addons':
        module_path = f'rdt.transformers.addons.{module_path.parent.name}.{module_name}'
    elif module_path.parent.parent.name == 'addons':
        module_path = (
            f'rdt.transformers.addons.{module_path.parent.parent.name}'
            f'.{module_path.parent.name}.{module_name}'
        )

    spec = importlib.util.spec_from_file_location(module_path, path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    return module


def validate_test_names(transformer):
    """Validate if the test methods are properly specified."""
    test_file = get_test_location(transformer)
    module = _load_module_from_path(test_file)

    test_class = getattr(module, f'Test{transformer.__name__}', None)
    assert test_class is not None, 'The expected test class was not found.'

    test_functions = inspect.getmembers(test_class, predicate=inspect.isfunction)
    test_functions = [
        test
        for test, _ in test_functions
        if test.startswith('test')
    ]

    assert test_functions, 'No test functions found within the test module.'

    transformer_functions = [
        name
        for name, function in transformer.__dict__.items()
        if isinstance(function, (FunctionType, classmethod, staticmethod))
    ]

    valid_test_functions = []
    for test in test_functions:
        count = len(valid_test_functions)
        for transformer_function in transformer_functions:
            simple_test = fr'test_{transformer_function}'
            described_test = fr'test_{transformer_function}_'
            if test.startswith(described_test):
                valid_test_functions.append(test)
            elif test.startswith(simple_test):
                valid_test_functions.append(test)

        fail_message = f'No function name was found for the test: {test}'
        assert len(valid_test_functions) > count, fail_message


@pytest.mark.parametrize('transformer', TRANSFORMERS.values(), ids=TRANSFORMERS.keys())  # noqa
def test_transformer_code_style(transformer):
    """Validate a transformer."""
    if not inspect.isclass(transformer):
        transformer = get_transformer_class(transformer)

    validate_transformer_subclass(transformer)
    validate_transformer_module(transformer)
    validate_test_location(transformer)
    validate_test_names(transformer)
    validate_transformer_addon(transformer)
    validate_transformer_importable_from_parent_module(transformer)



import numpy as np
import pandas as pd

from rdt.performance.datasets import numerical


class TestRandomIntegerGenerator:

    def test(self):
        output = numerical.RandomIntegerGenerator.generate(10)
        assert len(output) == 10
        assert output.dtype == int
        assert len(pd.unique(output)) > 1
        assert np.isnan(output).sum() == 0


class TestRandomIntegerNaNsGenerator:

    def test(self):
        output = numerical.RandomIntegerNaNsGenerator.generate(10)
        assert len(output) == 10
        assert output.dtype == float
        assert len(pd.unique(output)) > 1
        assert np.isnan(output).sum() > 0


class TestConstantIntegerGenerator:

    def test(self):
        output = numerical.ConstantIntegerGenerator.generate(10)
        assert len(output) == 10
        assert output.dtype == int
        assert len(pd.unique(output)) == 1
        assert np.isnan(output).sum() == 0


class TestConstantIntegerNaNsGenerator:

    def test(self):
        output = numerical.ConstantIntegerNaNsGenerator.generate(10)
        assert len(output) == 10
        assert output.dtype == float
        assert len(pd.unique(output)) == 2
        assert np.isnan(output).sum() >= 1


class TestAlmostConstantIntegerGenerator:

    def test(self):
        output = numerical.AlmostConstantIntegerGenerator.generate(10)
        assert len(output) == 10
        assert output.dtype == int
        assert len(pd.unique(output)) == 2
        assert np.isnan(output).sum() == 0


class TestAlmostConstantIntegerNaNsGenerator:

    def test(self):
        output = numerical.AlmostConstantIntegerNaNsGenerator.generate(10)
        assert len(output) == 10
        assert output.dtype == float
        assert len(pd.unique(output)) == 3
        assert np.isnan(output).sum() >= 1


class TestNormalGenerator:

    def test(self):
        output = numerical.NormalGenerator.generate(10)
        assert len(output) == 10
        assert output.dtype == float
        assert len(pd.unique(output)) == 10
        assert np.isnan(output).sum() == 0


class TestNormalNaNsGenerator:

    def test(self):
        output = numerical.NormalNaNsGenerator.generate(10)
        assert len(output) == 10
        assert output.dtype == float
        assert 1 < len(pd.unique(output)) <= 10
        assert np.isnan(output).sum() >= 1


class TestBigNormalGenerator:

    def test(self):
        output = numerical.BigNormalGenerator.generate(10)
        assert len(output) == 10
        assert output.dtype == float
        assert len(pd.unique(output)) == 10
        assert np.isnan(output).sum() == 0


class TestBigNormalNaNsGenerator:

    def test(self):
        output = numerical.BigNormalNaNsGenerator.generate(10)
        assert len(output) == 10
        assert output.dtype == float
        assert 1 < len(pd.unique(output)) <= 10
        assert np.isnan(output).sum() >= 1



"""Tests for the datasets.utils module."""

import numpy as np

from rdt.performance.datasets import utils


def test_add_nulls_int():
    array = np.arange(100)

    with_nans = utils.add_nans(array)

    assert len(with_nans) == 100
    assert 1 <= np.isnan(with_nans).sum() < 100

    nans = np.isnan(with_nans)
    np.testing.assert_array_equal(array[~nans], with_nans[~nans])


def test_add_nulls_float():
    array = np.arange(100).astype(float)

    with_nans = utils.add_nans(array)

    assert len(with_nans) == 100
    assert 1 <= np.isnan(with_nans).sum() < 100

    nans = np.isnan(with_nans)
    np.testing.assert_array_equal(array[~nans], with_nans[~nans])



"""IdentityTransformer module."""

from rdt.transformers.base import BaseTransformer


class IdentityTransformer(BaseTransformer):
    """Identity transformer that produces the same data.

    This transformer is intended for testing purposes only. The transform and reverse transform
    of this data is equal to the input.
    """

    def _fit(self, data):
        """Fit the transformer to the data.

        Args:
            data (pandas.Series or numpy.ndarray):
                Data to fit the transformer to.
        """
        self.OUTPUT_SDTYPES = {
            column: None
            for column in self.columns
        }

    def _transform(self, data):
        """Return the same input data.

        Args:
            data (pandas.Series or numpy.ndarray):
                Data to transform.

        Returns:
            pandas.DataFrame or pandas.Series
        """
        return data

    def _reverse_transform(self, data):
        """Return the same input data.

        Args:
            data (pandas.Series or numpy.ndarray):
                Data to revert.

        Returns:
            pandas.DataFrame or pandas.Series
        """
        return data



"""Unit tests for the NullTransformer."""

from unittest.mock import patch

import numpy as np
import pandas as pd

from rdt.transformers import NullTransformer


class TestNullTransformer:

    def test___init__default(self):
        """Test the initialization without passing any default arguments.

        When no arguments are passed, the attributes should be populated
        with the right values.

        Input:
            - nothing

        Expected Side Effects:
            - The `_missing_value_replacement` attribute should be `None`.
            - The `_model_missing_values` attribute should be `False`.
        """
        # Run
        transformer = NullTransformer()

        # Assert
        assert transformer._missing_value_replacement is None
        assert not transformer._model_missing_values

    def test___init__not_default(self):
        """Test the initialization passing values different than defaults.

        When arguments are passed, the attributes should be populated
        with the right values.

        Input:
            - Values different than the defaults.

        Expected Side Effects:
            - The attributes should be populated with the given values.
        """
        # Run
        transformer = NullTransformer('a_missing_value_replacement', False)

        # Assert
        assert transformer._missing_value_replacement == 'a_missing_value_replacement'
        assert not transformer._model_missing_values

    def test_models_missing_values(self):
        """Test the models_missing_values method.

        If the `model_missing_values` attributes evalutes to True, the
        `create_model_missing_values` method should return the same value.

        Setup:
            - Create an instance and set _model_missing_values to True

        Expected Output:
            - True
        """
        # Setup
        transformer = NullTransformer('something', model_missing_values=True)
        transformer._model_missing_values = True

        # Run
        models_missing_values = transformer.models_missing_values()

        # Assert
        assert models_missing_values

    def test__get_missing_value_replacement_scalar(self):
        """Test _get_missing_value_replacement when a scalar value is passed.

        If a missing_value_replacement different from None, 'mean' or 'mode' is
        passed to __init__, that value is returned.

        Setup:
            - NullTransformer passing a specific missing_value_replacement
              that is not None, mean or mode.

        Input:
            - A Series with some values.
            - A np.array with boolean values.

        Expected Output:
            - The value passed to __init__
        """
        # Setup
        transformer = NullTransformer('a_missing_value_replacement')

        # Run
        data = pd.Series([1, np.nan, 3], name='abc')
        missing_value_replacement = transformer._get_missing_value_replacement(data)

        # Assert
        assert missing_value_replacement == 'a_missing_value_replacement'

    def test__get_missing_value_replacement_all_nulls(self):
        """Test _get_missing_value_replacement when all the values are null.

        If the missing_value_replacement is not a scalar value and all the data
        values are null, the output be the mean, which is `np.nan`.

        Setup:
            - NullTransformer passing 'mean' as the missing_value_replacement.

        Input:
            - A Series filled with nan values.
            - A np.array of all True values.

        Expected Output:
            - 0
        """
        # Setup
        transformer = NullTransformer('mean')

        # Run
        data = pd.Series([np.nan, np.nan, np.nan], name='abc')
        missing_value_replacement = transformer._get_missing_value_replacement(data)

        # Assert
        assert missing_value_replacement is np.nan

    def test__get_missing_value_replacement_none_numerical(self):
        """Test _get_missing_value_replacement when missing_value_replacement is None.

        If the missing_value_replacement is None and the data is numerical,
        the output fill value should be the mean of the input data.

        Setup:
            - NullTransformer passing with default arguments.

        Input:
            - An Series filled with integer values such that the mean
              is not contained in the series and there is at least one null.
            - A np.array of booleans indicating which values are null.

        Expected Output:
            - The mean of the inputted Series.
        """
        # Setup
        transformer = NullTransformer('mean')

        # Run
        data = pd.Series([1, 2, np.nan], name='abc')
        missing_value_replacement = transformer._get_missing_value_replacement(data)

        # Assert
        assert missing_value_replacement == 1.5

    def test__get_missing_value_replacement_none_not_numerical(self):
        """Test _get_missing_value_replacement when missing_value_replacement is None.

        If the missing_value_replacement is None and the data is not numerical,
        the output fill value should be the mode of the input data.

        Setup:
            - NullTransformer with default arguments.

        Input:
            - An Series filled with string values with variable frequency and
              at least one null value.
            - A np.array of booleans indicating which values are null.

        Expected Output:
            - The most frequent value in the input series.
        """
        # Setup
        transformer = NullTransformer('mode')

        # Run
        data = pd.Series(['a', 'b', 'b', np.nan], name='abc')
        missing_value_replacement = transformer._get_missing_value_replacement(data)

        # Assert
        assert missing_value_replacement == 'b'

    def test__get_missing_value_replacement_mean(self):
        """Test _get_missing_value_replacement when missing_value_replacement is mean.

        If the missing_value_replacement is mean the output fill value should be the
        mean of the input data.

        Setup:
            - NullTransformer passing 'mean' as the missing_value_replacement.

        Input:
            - A Series filled with integer values such that the mean
              is not contained in the series and there is at least one null.
            - A np.array of booleans indicating which values are null.

        Expected Output:
            - The mode of the inputted Series.
        """
        # Setup
        transformer = NullTransformer('mean')

        # Run
        data = pd.Series([1, 2, np.nan], name='abc')
        missing_value_replacement = transformer._get_missing_value_replacement(data)

        # Assert
        assert missing_value_replacement == 1.5

    def test__get_missing_value_replacement_mode(self):
        """Test _get_missing_value_replacement when missing_value_replacement is 'mode'.

        If the missing_value_replacement is 'mode' the output fill value should be the
        mode of the input data.

        Setup:
            - NullTransformer passing 'mode' as the missing_value_replacement.

        Input:
            - A Series filled with integer values such that the mean
              is not contained in the series and there is at least one null.
            - A np.array of booleans indicating which values are null.

        Expected Output:
            - The most frequent value in the input series.
        """
        # Setup
        transformer = NullTransformer('mode')

        # Run
        data = pd.Series([1, 2, 2, np.nan], name='abc')
        missing_value_replacement = transformer._get_missing_value_replacement(data)

        # Assert
        assert missing_value_replacement == 2

    def test_fit_model_missing_values_none_and_nulls(self):
        """Test fit when null column is none and there are nulls.

        If there are nulls in the data and model_missing_values was given as None,
        then the _model_missing_values attribute should be set to True.
        Also validate that the null attribute and the _missing_value_replacement attributes
        are set accordingly.

        Setup:
            - A NullTransformer with default arguments.

        Input:
            - pd.Series of integers that contains nulls.

        Expected Side Effects:
            - the model_missing_values attribute should be set to True.
            - the nulls attribute should be set to True.
            - the missing_value_replacement should be set to the mean of the given integers.
        """
        # Setup
        transformer = NullTransformer(missing_value_replacement='mean', model_missing_values=True)

        # Run
        data = pd.Series([1, 2, np.nan])
        transformer.fit(data)

        # Assert
        assert transformer.nulls
        assert transformer._model_missing_values
        assert transformer._missing_value_replacement == 1.5

    def test_fit_model_missing_values_none_and_no_nulls(self):
        """Test fit when null column is none and there are NO nulls.

        If there are no nulls in the data and model_missing_values was given as ``False``,
        then the _model_missing_values attribute should be set to ``False``.
        Also validate that the null attribute and the ``_missing_value_replacement`` attributes
        are set accordingly.

        Setup:
            - A NullTransformer with default arguments.

        Input:
            - pd.Series of strings that contains no nulls.

        Expected Side Effects:
            - the model_missing_values attribute should be set to False.
            - the nulls attribute should be set to False.
            - the missing_value_replacement should be set to ``np.nan``, default.
        """
        # Setup
        transformer = NullTransformer()

        # Run
        data = pd.Series(['a', 'b', 'b'])
        transformer.fit(data)

        # Assert
        assert not transformer.nulls
        assert not transformer._model_missing_values
        assert transformer._missing_value_replacement is None

    def test_fit_model_missing_values_not_none(self):
        """Test fit when null column is set to True/False.

        If model_missing_values is set to True or False, the _model_missing_values should
        get that value regardless of whether there are nulls or not.

        Notice that this test covers 4 scenarios at once.

        Setup:
            - 4 NullTransformer intances, 2 of them passing False for the model_missing_values
              and 2 of them passing True.

        Input:
            - 2 pd.Series, one containing nulls and the other not containing nulls.

        Expected Side Effects:
            - the _model_missing_values attribute should be set to True or False as indicated
              in the Transformer creation.
            - the nulls attribute should be True or False depending on whether
              the input data contains nulls or not.
        """
        # Setup
        model_missing_values_false_nulls = NullTransformer(
            missing_value_replacement='mode',
            model_missing_values=False
        )
        model_missing_values_false_no_nulls = NullTransformer(
            missing_value_replacement='mode',
            model_missing_values=False
        )
        model_missing_values_true_nulls = NullTransformer(
            missing_value_replacement='mean',
            model_missing_values=True
        )
        model_missing_values_true_no_nulls = NullTransformer(
            missing_value_replacement='mean',
            model_missing_values=True
        )
        nulls_str = pd.Series(['a', 'b', 'b', np.nan])
        no_nulls_str = pd.Series(['a', 'b', 'b', 'c'])
        nulls_int = pd.Series([1, 2, 3, np.nan])
        no_nulls_int = pd.Series([1, 2, 3, 4])

        # Run
        model_missing_values_false_nulls.fit(nulls_str)
        model_missing_values_false_no_nulls.fit(no_nulls_str)
        model_missing_values_true_nulls.fit(nulls_int)
        model_missing_values_true_no_nulls.fit(no_nulls_int)

        # Assert
        assert not model_missing_values_false_nulls._model_missing_values
        assert model_missing_values_false_nulls.nulls
        assert model_missing_values_false_nulls._missing_value_replacement == 'b'

        assert not model_missing_values_false_no_nulls._model_missing_values
        assert not model_missing_values_false_no_nulls.nulls
        assert model_missing_values_false_no_nulls._missing_value_replacement == 'b'

        assert model_missing_values_true_nulls._model_missing_values
        assert model_missing_values_true_nulls.nulls
        assert model_missing_values_true_nulls._missing_value_replacement == 2

        assert not model_missing_values_true_no_nulls._model_missing_values
        assert not model_missing_values_true_no_nulls.nulls
        assert model_missing_values_true_no_nulls._missing_value_replacement == 2.5

    def test_transform__model_missing_values_true(self):
        """Test transform when _model_missing_values.

        When _model_missing_values, the nulls should be replaced
        by the _missing_value_replacement and another column flagging the nulls
        should be created.

        Setup:
            - NullTransformer instance with _model_missing_values set to True,
              _missing_value_replacement set to a scalar value.

        Input:
            - A pd.Series of strings with nulls.

        Expected Output:
            - Exactly the same as the input, replacing the nulls with the
              scalar value.

        Expected Side Effects:
            - The input data has the null values replaced.
        """
        # Setup
        transformer = NullTransformer()
        transformer.nulls = False
        transformer._model_missing_values = True
        transformer._missing_value_replacement = 'c'
        input_data = pd.Series(['a', 'b', np.nan])

        # Run
        output = transformer.transform(input_data)

        # Assert
        expected_output = np.array([
            ['a', 0.0],
            ['b', 0.0],
            ['c', 1.0],
        ], dtype=object)
        np.testing.assert_equal(expected_output, output)

    def test_transform__model_missing_values_false(self):
        """Test transform when _model_missing_values is False.

        When the _model_missing_values is false, the nulls should be replaced
        by the _missing_value_replacement.

        Setup:
            - NullTransformer instance with _model_missing_values set to False,
              _missing_value_replacement set to a scalar value.

        Input:
            - A pd.Series of integers with nulls.

        Expected Output:
            - Same data as the input, replacing the nulls with the
              scalar value.

        Expected Side Effects:
            - The input data has not been modified.
        """
        # Setup
        transformer = NullTransformer()
        transformer._model_missing_values = False
        transformer._missing_value_replacement = 3
        input_data = pd.Series([1, 2, np.nan])

        # Run
        output = transformer.transform(input_data)

        # Assert
        expected_output = np.array([1, 2, 3])
        np.testing.assert_equal(expected_output, output)

        modified_input_data = pd.Series([1, 2, np.nan])
        pd.testing.assert_series_equal(modified_input_data, input_data)

    def test_reverse_transform__model_missing_values_true_nulls_true(self):
        """Test reverse_transform when _model_missing_values and nulls are True.

        When _model_missing_values and nulls attributes are both True, the second column
        in the input data should be used to decide which values to replace
        with nan, by selecting the rows where the null column value is > 0.5.

        Setup:
            - NullTransformer instance with _model_missing_values and nulls
              attributes set to True.

        Input:
            - 2d numpy array with variate float values.

        Expected Output:
            - pd.Series containing the first column from the input data
              with the values indicated by the first column replaced by nans.

        Expected Side Effects:
            - the input data should have been modified.
        """
        # Setup
        transformer = NullTransformer()
        transformer._model_missing_values = True
        transformer.nulls = True
        input_data = np.array([
            [0.0, 0.0],
            [0.2, 0.2],
            [0.4, 0.4],
            [0.6, 0.6],
            [0.8, 0.8],
        ])

        # Run
        output = transformer.reverse_transform(input_data)

        # Assert
        expected_output = pd.Series([0.0, 0.2, 0.4, np.nan, np.nan])
        pd.testing.assert_series_equal(expected_output, output)

    def test_reverse_transform__model_missing_values_true_nulls_false(self):
        """Test reverse_transform when _model_missing_values and nulls is False.

        When _model_missing_values but nulls are False, the second column of the
        input data must be dropped and the first one returned as a Series without
        having been modified.

        Setup:
            - NullTransformer instance with _model_missing_values set to True and nulls
              attribute set to False

        Input:
            - 2d numpy array with variate float values.

        Expected Output:
            - pd.Series containing the first column from the input data unmodified.
        """
        # Setup
        transformer = NullTransformer()
        transformer._model_missing_values = True
        transformer.nulls = False
        input_data = np.array([
            [0.0, 0.0],
            [0.2, 0.2],
            [0.4, 0.4],
            [0.6, 0.6],
            [0.8, 0.8],
        ])

        # Run
        output = transformer.reverse_transform(input_data)

        # Assert
        expected_output = pd.Series([0.0, 0.2, 0.4, 0.6, 0.8])
        pd.testing.assert_series_equal(expected_output, output)

    @patch('rdt.transformers.null.np.random')
    def test_reverse_transform__model_missing_values_false_nulls_true(self, random_mock):
        """Test reverse_transform when _model_missing_values is False and nulls.

        When _model_missing_values is False and the nulls attribute, a ``_null_percentage``
        of values should randomly be replaced with ``np.nan``.

        Setup:
            - NullTransformer instance with _model_missing_values set to False and nulls
              attribute set to True.
            - A mock for ``np.random``.

        Input:
            - 1d numpy array with variate float values.

        Expected Output:
            - pd.Series containing the same data as input, with the random values
            replaced with ``np.nan``.
        """
        # Setup
        transformer = NullTransformer()
        transformer._model_missing_values = False
        transformer.nulls = True
        transformer._null_percentage = 0.5
        input_data = np.array([0.0, 0.2, 0.4, 0.6])
        random_mock.random.return_value = np.array([1, 1, 0, 1])

        # Run
        output = transformer.reverse_transform(input_data)

        # Assert
        expected_output = pd.Series([0.0, 0.2, np.nan, 0.6])
        pd.testing.assert_series_equal(expected_output, output)

    def test_reverse_transform__model_missing_values_false_nulls_false(self):
        """Test reverse_transform when _model_missing_values is False and nulls.

        When _model_missing_values is False and the nulls attribute is also False, the
        input data is not modified at all.

        Setup:
            - NullTransformer instance with _model_missing_values and nulls attributes
              set to False, and the _missing_value_replacement set to a scalar value.

        Input:
            - 1d numpy array with variate float values, containing the _missing_value_replacement
              among them.

        Expected Output:
            - pd.Series containing the same data as input, without modification.
        """
        # Setup
        transformer = NullTransformer()
        transformer._model_missing_values = False
        transformer.nulls = False
        transformer._missing_value_replacement = 0.4
        input_data = np.array([0.0, 0.2, 0.4, 0.6])

        # Run
        output = transformer.reverse_transform(input_data)

        # Assert
        expected_output = pd.Series([0.0, 0.2, 0.4, 0.6])
        pd.testing.assert_series_equal(expected_output, output)



import re
from unittest.mock import Mock, call, patch

import numpy as np
import pandas as pd
import pytest

from rdt.errors import Error
from rdt.transformers.categorical import (
    CustomLabelEncoder, FrequencyEncoder, LabelEncoder, OneHotEncoder)

RE_SSN = re.compile(r'\d\d\d-\d\d-\d\d\d\d')


class TestFrequencyEncoder:

    def test___setstate__(self):
        """Test the ``__set_state__`` method.

        Validate that the ``__dict__`` attribute is correctly udpdated when

        Setup:
            - create an instance of a ``FrequencyEncoder``.

        Side effect:
            - it updates the ``__dict__`` attribute of the object.
        """
        # Setup
        transformer = FrequencyEncoder()

        # Run
        transformer.__setstate__({
            'intervals': {
                None: 'abc'
            }
        })

        # Assert
        assert transformer.__dict__['intervals'][np.nan] == 'abc'

    def test___init__(self):
        """Passed arguments must be stored as attributes."""
        # Run
        transformer = FrequencyEncoder(add_noise='add_noise_value')

        # Asserts
        assert transformer.add_noise == 'add_noise_value'

    def test_is_transform_deterministic(self):
        """Test the ``is_transform_deterministic`` method.

        Validate the method returns the opposite boolean value of the ``add_noise`` parameter.

        Setup:
            - initialize a ``FrequencyEncoder`` with ``add_noise = True``.

        Output:
            - the boolean value which is the opposite of ``add_noise``.
        """
        # Setup
        transformer = FrequencyEncoder(add_noise=True)

        # Run
        output = transformer.is_transform_deterministic()

        # Assert
        assert output is False

    def test_is_composition_identity(self):
        """Test the ``is_composition_identity`` method.

        Since ``COMPOSITION_IS_IDENTITY`` is True, just validates that the method
        returns the opposite boolean value of the ``add_noise`` parameter.

        Setup:
            - initialize a ``FrequencyEncoder`` with ``add_noise = True``.

        Output:
            - the boolean value which is the opposite of ``add_noise``.
        """
        # Setup
        transformer = FrequencyEncoder(add_noise=True)

        # Run
        output = transformer.is_composition_identity()

        # Assert
        assert output is False

    def test__get_intervals(self):
        """Test the ``_get_intervals`` method.

        Validate that the intervals for each categorical value are correct.

        Input:
            - a pandas series containing categorical values.

        Output:
            - a tuple, where the first element describes the intervals for each
            categorical value (start, end).
        """
        # Run
        data = pd.Series(['foo', 'bar', 'bar', 'foo', 'foo', 'tar'])
        result = FrequencyEncoder._get_intervals(data)

        # Asserts
        expected_intervals = {
            'foo': (
                0,
                0.5,
                0.25,
                0.5 / 6
            ),
            'bar': (
                0.5,
                0.8333333333333333,
                0.6666666666666666,
                0.05555555555555555
            ),
            'tar': (
                0.8333333333333333,
                0.9999999999999999,
                0.9166666666666666,
                0.027777777777777776
            )
        }
        expected_means = pd.Series({
            'foo': 0.25,
            'bar': 0.6666666666666666,
            'tar': 0.9166666666666666
        })
        expected_starts = pd.DataFrame({
            'category': ['foo', 'bar', 'tar'],
            'start': [0, 0.5, 0.8333333333333333]
        }).set_index('start')

        assert result[0] == expected_intervals
        pd.testing.assert_series_equal(result[1], expected_means)
        pd.testing.assert_frame_equal(result[2], expected_starts)

    def test__get_intervals_nans(self):
        """Test the ``_get_intervals`` method when data contains nan's.

        Validate that the intervals for each categorical value are correct, when passed
        data containing nan values.

        Input:
            - a pandas series cotaining nan values and categorical values.

        Output:
            - a tuple, where the first element describes the intervals for each
            categorical value (start, end).
        """
        # Setup
        data = pd.Series(['foo', np.nan, None, 'foo', 'foo', 'tar'])

        # Run
        result = FrequencyEncoder._get_intervals(data)

        # Assert
        expected_intervals = {
            'foo': (
                0,
                0.5,
                0.25,
                0.5 / 6
            ),
            np.nan: (
                0.5,
                0.8333333333333333,
                0.6666666666666666,
                0.05555555555555555
            ),
            'tar': (
                0.8333333333333333,
                0.9999999999999999,
                0.9166666666666666,
                0.027777777777777776
            )
        }
        expected_means = pd.Series({
            'foo': 0.25,
            np.nan: 0.6666666666666666,
            'tar': 0.9166666666666666
        })
        expected_starts = pd.DataFrame({
            'category': ['foo', np.nan, 'tar'],
            'start': [0, 0.5, 0.8333333333333333]
        }).set_index('start')

        assert result[0] == expected_intervals
        pd.testing.assert_series_equal(result[1], expected_means)
        pd.testing.assert_frame_equal(result[2], expected_starts)

    def test__fit_intervals(self):
        # Setup
        transformer = FrequencyEncoder()

        # Run
        data = pd.Series(['foo', 'bar', 'bar', 'foo', 'foo', 'tar'])
        transformer._fit(data)

        # Asserts
        expected_intervals = {
            'foo': (
                0,
                0.5,
                0.25,
                0.5 / 6
            ),
            'bar': (
                0.5,
                0.8333333333333333,
                0.6666666666666666,
                0.05555555555555555
            ),
            'tar': (
                0.8333333333333333,
                0.9999999999999999,
                0.9166666666666666,
                0.027777777777777776
            )
        }
        expected_means = pd.Series({
            'foo': 0.25,
            'bar': 0.6666666666666666,
            'tar': 0.9166666666666666
        })
        expected_starts = pd.DataFrame({
            'category': ['foo', 'bar', 'tar'],
            'start': [0, 0.5, 0.8333333333333333]
        }).set_index('start')

        assert transformer.intervals == expected_intervals
        pd.testing.assert_series_equal(transformer.means, expected_means)
        pd.testing.assert_frame_equal(transformer.starts, expected_starts)

    def test__get_value_add_noise_false(self):
        # Setup
        transformer = FrequencyEncoder(add_noise=False)
        transformer.intervals = {
            'foo': (0, 0.5, 0.25, 0.5 / 6),
            np.nan: (0.5, 1.0, 0.75, 0.5 / 6),
        }

        # Run
        result_foo = transformer._get_value('foo')
        result_nan = transformer._get_value(np.nan)

        # Asserts
        assert result_foo == 0.25
        assert result_nan == 0.75

    @patch('rdt.transformers.categorical.norm')
    def test__get_value_add_noise_true(self, norm_mock):
        # setup
        norm_mock.rvs.return_value = 0.2745

        transformer = FrequencyEncoder(add_noise=True)
        transformer.intervals = {
            'foo': (0, 0.5, 0.25, 0.5 / 6),
        }

        # Run
        result = transformer._get_value('foo')

        # Asserts
        assert result == 0.2745

    def test__reverse_transform_array(self):
        """Test reverse_transform a numpy.array"""
        # Setup
        data = pd.Series(['foo', 'bar', 'bar', 'foo', 'foo', 'tar'])
        rt_data = np.array([-0.6, 0.5, 0.6, 0.2, 0.1, -0.2])
        transformer = FrequencyEncoder()

        # Run
        transformer._fit(data)
        result = transformer._reverse_transform(rt_data)

        # Asserts
        expected_intervals = {
            'foo': (
                0,
                0.5,
                0.25,
                0.5 / 6
            ),
            'bar': (
                0.5,
                0.8333333333333333,
                0.6666666666666666,
                0.05555555555555555
            ),
            'tar': (
                0.8333333333333333,
                0.9999999999999999,
                0.9166666666666666,
                0.027777777777777776
            )
        }

        assert transformer.intervals == expected_intervals

        expect = pd.Series(['foo', 'bar', 'bar', 'foo', 'foo', 'foo'])
        pd.testing.assert_series_equal(result, expect)

    def test__transform_user_warning(self):
        """Test the ``_transform`` method generates the correct user warning.

        When asked to transform data not seen during the fit, a UserWarning should be raised.

        Setup:
            - create an instance of the ``FrequencyEncoder``, where ``means`` is a list
            of floats and ``intervals`` is the appropriate dictionary.

        Input:
            - a pandas series containing a np.nan.

        Output:
            - a numpy array containing the transformed data.

        Raises:
            - a UserWarning with the correct message.
        """
        # Setup
        data = pd.Series([1, 2, 3, 4, np.nan])
        transformer = FrequencyEncoder()
        transformer.means = [0.125, 0.375, 0.625, 0.875]
        transformer.intervals = {
            4: (0, 0.25, 0.125, 0.041666666666666664),
            3: (0.25, 0.5, 0.375, 0.041666666666666664),
            2: (0.5, 0.75, 0.625, 0.041666666666666664),
            1: (0.75, 1.0, 0.875, 0.041666666666666664),
        }

        # Run
        warning_msg = re.escape(
            'The data contains 1 new categories that were not '
            'seen in the original data (examples: {nan}). Assigning '
            'them random values. If you want to model new categories, '
            'please fit the transformer again with the new data.'
        )
        with pytest.warns(UserWarning, match=warning_msg):
            transformed = transformer._transform(data)

        # Assert
        expected = pd.Series([0.875, 0.625, 0.375, 0.125])
        np.testing.assert_array_equal(transformed[:4], expected)

        assert transformed[4] in transformer.means

    def test__transform_by_category_called(self):
        """Test that the `_transform_by_category` method is called.

        When the number of rows is greater than the number of categories, expect
        that the `_transform_by_category` method is called.

        Setup:
            The categorical transformer is instantiated with 4 categories.

        Input:
            - data with 5 rows.

        Output:
            - the output of `_transform_by_category`.

        Side effects:
            - `_transform_by_category` will be called once.
        """
        # Setup
        data = pd.Series([1, 3, 3, 2, 1])

        categorical_transformer_mock = Mock()
        categorical_transformer_mock.means = pd.Series([0.125, 0.375, 0.625, 0.875])

        # Run
        transformed = FrequencyEncoder._transform(categorical_transformer_mock, data)

        # Asserts
        categorical_transformer_mock._transform_by_category.assert_called_once_with(data)
        assert transformed == categorical_transformer_mock._transform_by_category.return_value

    def test__transform_by_category(self):
        """Test the `_transform_by_category` method with numerical data.

        Expect that the correct transformed data is returned.

        Setup:
            The categorical transformer is instantiated with 4 categories and intervals.

        Input:
            - data with 5 rows.

        Ouptut:
            - the transformed data.
        """
        # Setup
        data = pd.Series([1, 3, 3, 2, 1])
        transformer = FrequencyEncoder()
        transformer.intervals = {
            4: (0, 0.25, 0.125, 0.041666666666666664),
            3: (0.25, 0.5, 0.375, 0.041666666666666664),
            2: (0.5, 0.75, 0.625, 0.041666666666666664),
            1: (0.75, 1.0, 0.875, 0.041666666666666664),
        }

        # Run
        transformed = transformer._transform_by_category(data)

        # Asserts
        expected = np.array([0.875, 0.375, 0.375, 0.625, 0.875])
        assert (transformed == expected).all()

    def test__transform_by_category_nans(self):
        """Test the ``_transform_by_category`` method with data containing nans.

        Validate that the data is transformed correctly when it contains nan's.

        Setup:
            - the categorical transformer is instantiated, and the appropriate ``intervals``
            attribute is set.

        Input:
            - a pandas series containing nan's.

        Output:
            - a numpy array containing the transformed data.
        """
        # Setup
        data = pd.Series([np.nan, 3, 3, 2, np.nan])
        transformer = FrequencyEncoder()
        transformer.intervals = {
            4: (0, 0.25, 0.125, 0.041666666666666664),
            3: (0.25, 0.5, 0.375, 0.041666666666666664),
            2: (0.5, 0.75, 0.625, 0.041666666666666664),
            np.nan: (0.75, 1.0, 0.875, 0.041666666666666664),
        }

        # Run
        transformed = transformer._transform_by_category(data)

        # Asserts
        expected = np.array([0.875, 0.375, 0.375, 0.625, 0.875])
        assert (transformed == expected).all()

    @patch('rdt.transformers.categorical.norm')
    def test__transform_by_category_add_noise_true(self, norm_mock):
        """Test the ``_transform_by_category`` method when ``add_noise`` is True.

        Validate that the data is transformed correctly when ``add_noise`` is True.

        Setup:
            - the categorical transformer is instantiated with ``add_noise`` as True,
            and the appropriate ``intervals`` attribute is set.
            - the ``intervals`` attribute is set to a a dictionary of intervals corresponding
            to the elements of the passed data.
            - set the ``side_effect`` of the ``rvs_mock`` to the appropriate function.

        Input:
            - a pandas series.

        Output:
            - a numpy array containing the transformed data.

        Side effect:
            - ``rvs_mock`` should be called four times, one for each element of the
            intervals dictionary.
        """
        # Setup
        def rvs_mock_func(loc, scale, **kwargs):
            return loc

        norm_mock.rvs.side_effect = rvs_mock_func

        data = pd.Series([1, 3, 3, 2, 1])
        transformer = FrequencyEncoder(add_noise=True)
        transformer.intervals = {
            4: (0, 0.25, 0.125, 0.041666666666666664),
            3: (0.25, 0.5, 0.375, 0.041666666666666664),
            2: (0.5, 0.75, 0.625, 0.041666666666666664),
            1: (0.75, 1.0, 0.875, 0.041666666666666664),
        }

        # Run
        transformed = transformer._transform_by_category(data)

        # Assert
        expected = np.array([0.875, 0.375, 0.375, 0.625, 0.875])
        assert (transformed == expected).all()
        norm_mock.rvs.assert_has_calls([
            call(0.125, 0.041666666666666664, size=0),
            call(0.375, 0.041666666666666664, size=2),
            call(0.625, 0.041666666666666664, size=1),
            call(0.875, 0.041666666666666664, size=2),
        ])

    def test__transform_by_row_called(self):
        """Test that the `_transform_by_row` method is called.

        When the number of rows is less than or equal to the number of categories,
        expect that the `_transform_by_row` method is called.

        Setup:
            The categorical transformer is instantiated with 4 categories.
        Input:
            - data with 4 rows
        Output:
            - the output of `_transform_by_row`
        Side effects:
            - `_transform_by_row` will be called once
        """
        # Setup
        data = pd.Series([1, 2, 3, 4])

        categorical_transformer_mock = Mock()
        categorical_transformer_mock.means = pd.Series([0.125, 0.375, 0.625, 0.875])

        # Run
        transformed = FrequencyEncoder._transform(categorical_transformer_mock, data)

        # Asserts
        categorical_transformer_mock._transform_by_row.assert_called_once_with(data)
        assert transformed == categorical_transformer_mock._transform_by_row.return_value

    def test__transform_by_row(self):
        """Test the `_transform_by_row` method with numerical data.

        Expect that the correct transformed data is returned.

        Setup:
            The categorical transformer is instantiated with 4 categories and intervals.
        Input:
            - data with 4 rows
        Ouptut:
            - the transformed data
        """
        # Setup
        data = pd.Series([1, 2, 3, 4])
        transformer = FrequencyEncoder()
        transformer.intervals = {
            4: (0, 0.25, 0.125, 0.041666666666666664),
            3: (0.25, 0.5, 0.375, 0.041666666666666664),
            2: (0.5, 0.75, 0.625, 0.041666666666666664),
            1: (0.75, 1.0, 0.875, 0.041666666666666664),
        }

        # Run
        transformed = transformer._transform_by_row(data)

        # Asserts
        expected = np.array([0.875, 0.625, 0.375, 0.125])
        assert (transformed == expected).all()

    @patch('psutil.virtual_memory')
    def test__reverse_transform_by_matrix_called(self, psutil_mock):
        """Test that the `_reverse_transform_by_matrix` method is called.

        When there is enough virtual memory, expect that the
        `_reverse_transform_by_matrix` method is called.

        Setup:
            The categorical transformer is instantiated with 4 categories. Also patch the
            `psutil.virtual_memory` function to return a large enough `available_memory`.
        Input:
            - numerical data with 4 rows
        Output:
            - the output of `_reverse_transform_by_matrix`
        Side effects:
            - `_reverse_transform_by_matrix` will be called once
        """
        # Setup
        data = pd.Series([1, 2, 3, 4])

        categorical_transformer_mock = Mock()
        categorical_transformer_mock.means = pd.Series([0.125, 0.375, 0.625, 0.875])

        virtual_memory = Mock()
        virtual_memory.available = 4 * 4 * 8 * 3 + 1
        psutil_mock.return_value = virtual_memory

        # Run
        reverse = FrequencyEncoder._reverse_transform(categorical_transformer_mock, data)

        # Asserts
        reverse_arg = categorical_transformer_mock._reverse_transform_by_matrix.call_args[0][0]
        np.testing.assert_array_equal(reverse_arg, data.clip(0, 1))
        assert reverse == categorical_transformer_mock._reverse_transform_by_matrix.return_value

    @patch('psutil.virtual_memory')
    def test__reverse_transform_by_matrix(self, psutil_mock):
        """Test the _reverse_transform_by_matrix method with numerical data

        Expect that the transformed data is correctly reverse transformed.

        Setup:
            The categorical transformer is instantiated with 4 categories and means. Also patch
            the `psutil.virtual_memory` function to return a large enough `available_memory`.
        Input:
            - transformed data with 4 rows
        Ouptut:
            - the original data
        """
        # Setup
        data = pd.Series([1, 2, 3, 4])
        transformed = pd.Series([0.875, 0.625, 0.375, 0.125])

        transformer = FrequencyEncoder()
        transformer.means = pd.Series([0.125, 0.375, 0.625, 0.875], index=[4, 3, 2, 1])
        transformer.dtype = data.dtype

        virtual_memory = Mock()
        virtual_memory.available = 4 * 4 * 8 * 3 + 1
        psutil_mock.return_value = virtual_memory

        # Run
        reverse = transformer._reverse_transform_by_matrix(transformed)

        # Assert
        pd.testing.assert_series_equal(data, reverse)

    @patch('psutil.virtual_memory')
    def test__reverse_transform_by_category_called(self, psutil_mock):
        """Test that the `_reverse_transform_by_category` method is called.

        When there is not enough virtual memory and the number of rows is greater than the
        number of categories, expect that the `_reverse_transform_by_category` method is called.

        Setup:
            The categorical transformer is instantiated with 4 categories. Also patch the
            `psutil.virtual_memory` function to return an `available_memory` of 1.
        Input:
            - numerical data with 5 rows
        Output:
            - the output of `_reverse_transform_by_category`
        Side effects:
            - `_reverse_transform_by_category` will be called once
        """
        # Setup
        transform_data = pd.Series([1, 3, 3, 2, 1])

        categorical_transformer_mock = Mock()
        categorical_transformer_mock.means = pd.Series([0.125, 0.375, 0.625, 0.875])

        virtual_memory = Mock()
        virtual_memory.available = 1
        psutil_mock.return_value = virtual_memory

        # Run
        reverse = FrequencyEncoder._reverse_transform(
            categorical_transformer_mock, transform_data)

        # Asserts
        reverse_arg = categorical_transformer_mock._reverse_transform_by_category.call_args[0][0]
        np.testing.assert_array_equal(reverse_arg, transform_data.clip(0, 1))
        assert reverse == categorical_transformer_mock._reverse_transform_by_category.return_value

    @patch('psutil.virtual_memory')
    def test__reverse_transform_by_category(self, psutil_mock):
        """Test the _reverse_transform_by_category method with numerical data.

        Expect that the transformed data is correctly reverse transformed.

        Setup:
            The categorical transformer is instantiated with 4 categories, and the means
            and intervals are set for those categories. Also patch the `psutil.virtual_memory`
            function to return an `available_memory` of 1.
        Input:
            - transformed data with 5 rows
        Ouptut:
            - the original data
        """
        data = pd.Series([1, 3, 3, 2, 1])
        transformed = pd.Series([0.875, 0.375, 0.375, 0.625, 0.875])

        transformer = FrequencyEncoder()
        transformer.means = pd.Series([0.125, 0.375, 0.625, 0.875], index=[4, 3, 2, 1])
        transformer.intervals = {
            4: (0, 0.25, 0.125, 0.041666666666666664),
            3: (0.25, 0.5, 0.375, 0.041666666666666664),
            2: (0.5, 0.75, 0.625, 0.041666666666666664),
            1: (0.75, 1.0, 0.875, 0.041666666666666664),
        }
        transformer.dtype = data.dtype

        virtual_memory = Mock()
        virtual_memory.available = 1
        psutil_mock.return_value = virtual_memory

        reverse = transformer._reverse_transform_by_category(transformed)

        pd.testing.assert_series_equal(data, reverse)

    def test__get_category_from_start(self):
        """Test the ``_get_category_from_start`` method.

        Setup:
            - instantiate a ``FrequencyEncoder``, and set the attribute ``starts``
            to a pandas dataframe with ``set_index`` as ``'start'``.

        Input:
            - an integer, an index from data.

        Output:
            - a category from the data.
        """
        # Setup
        transformer = FrequencyEncoder()
        transformer.starts = pd.DataFrame({
            'start': [0.0, 0.5, 0.7],
            'category': ['a', 'b', 'c']
        }).set_index('start')

        # Run
        category = transformer._get_category_from_start(2)

        # Assert
        assert category == 'c'

    @patch('psutil.virtual_memory')
    def test__reverse_transform_by_row_called(self, psutil_mock):
        """Test that the `_reverse_transform_by_row` method is called.

        When there is not enough virtual memory and the number of rows is less than or equal
        to the number of categories, expect that the `_reverse_transform_by_row` method
        is called.

        Setup:
            The categorical transformer is instantiated with 4 categories. Also patch the
            `psutil.virtual_memory` function to return an `available_memory` of 1.
        Input:
            - numerical data with 4 rows
        Output:
            - the output of `_reverse_transform_by_row`
        Side effects:
            - `_reverse_transform_by_row` will be called once
        """
        # Setup
        data = pd.Series([1, 2, 3, 4])

        categorical_transformer_mock = Mock()
        categorical_transformer_mock.means = pd.Series([0.125, 0.375, 0.625, 0.875])
        categorical_transformer_mock.starts = pd.DataFrame(
            [0., 0.25, 0.5, 0.75], index=[4, 3, 2, 1], columns=['category'])
        categorical_transformer_mock._normalize.return_value = data

        virtual_memory = Mock()
        virtual_memory.available = 1
        psutil_mock.return_value = virtual_memory

        # Run
        reverse = FrequencyEncoder._reverse_transform(categorical_transformer_mock, data)

        # Asserts
        reverse_arg = categorical_transformer_mock._reverse_transform_by_row.call_args[0][0]
        np.testing.assert_array_equal(reverse_arg, data.clip(0, 1))
        assert reverse == categorical_transformer_mock._reverse_transform_by_row.return_value

    @patch('psutil.virtual_memory')
    def test__reverse_transform_by_row(self, psutil_mock):
        """Test the _reverse_transform_by_row method with numerical data.

        Expect that the transformed data is correctly reverse transformed.

        Setup:
            The categorical transformer is instantiated with 4 categories, and the means, starts,
            and intervals are set for those categories. Also patch the `psutil.virtual_memory`
            function to return an `available_memory` of 1.
        Input:
            - transformed data with 4 rows
        Ouptut:
            - the original data
        """
        # Setup
        data = pd.Series([1, 2, 3, 4])
        transformed = pd.Series([0.875, 0.625, 0.375, 0.125])

        transformer = FrequencyEncoder()
        transformer.means = pd.Series([0.125, 0.375, 0.625, 0.875], index=[4, 3, 2, 1])
        transformer.starts = pd.DataFrame(
            [4, 3, 2, 1], index=[0., 0.25, 0.5, 0.75], columns=['category'])
        transformer.intervals = {
            4: (0, 0.25, 0.125, 0.041666666666666664),
            3: (0.25, 0.5, 0.375, 0.041666666666666664),
            2: (0.5, 0.75, 0.625, 0.041666666666666664),
            1: (0.75, 1.0, 0.875, 0.041666666666666664),
        }
        transformer.dtype = data.dtype

        virtual_memory = Mock()
        virtual_memory.available = 1
        psutil_mock.return_value = virtual_memory

        # Run
        reverse = transformer._reverse_transform(transformed)

        # Assert
        pd.testing.assert_series_equal(data, reverse)


class TestOneHotEncoder:

    def test__prepare_data_empty_lists(self):
        # Setup
        ohe = OneHotEncoder()
        data = [[], [], []]

        # Assert
        with pytest.raises(ValueError, match='Unexpected format.'):
            ohe._prepare_data(data)

    def test__prepare_data_nested_lists(self):
        # Setup
        ohe = OneHotEncoder()
        data = [[[]]]

        # Assert
        with pytest.raises(ValueError, match='Unexpected format.'):
            ohe._prepare_data(data)

    def test__prepare_data_list_of_lists(self):
        # Setup
        ohe = OneHotEncoder()

        # Run
        data = [['a'], ['b'], ['c']]
        out = ohe._prepare_data(data)

        # Assert
        expected = np.array(['a', 'b', 'c'])
        np.testing.assert_array_equal(out, expected)

    def test__prepare_data_pandas_series(self):
        # Setup
        ohe = OneHotEncoder()

        # Run
        data = pd.Series(['a', 'b', 'c'])
        out = ohe._prepare_data(data)

        # Assert
        expected = pd.Series(['a', 'b', 'c'])
        np.testing.assert_array_equal(out, expected)

    def test_get_output_sdtypes(self):
        """Test the ``get_output_sdtypes`` method.

        Validate that the ``_add_prefix`` method is properly applied to the ``output_sdtypes``
        dictionary. For this class, the ``output_sdtypes`` dictionary is described as:

        {
            'value1': 'float',
            'value2': 'float',
            ...
        }

        The number of items in the dictionary is defined by the ``dummies`` attribute.

        Setup:
            - initialize a ``OneHotEncoder`` and set:
                - the ``dummies`` attribute to a list.
                - the ``column_prefix`` attribute to a string.

        Output:
            - the ``output_sdtypes`` dictionary, but with ``self.column_prefix``
            added to the beginning of the keys of the ``output_sdtypes`` dictionary.
        """
        # Setup
        transformer = OneHotEncoder()
        transformer.column_prefix = 'abc'
        transformer.dummies = [1, 2]

        # Run
        output = transformer.get_output_sdtypes()

        # Assert
        expected = {
            'abc.value0': 'float',
            'abc.value1': 'float'
        }
        assert output == expected

    def test__fit_dummies_no_nans(self):
        """Test the ``_fit`` method without nans.

        Check that ``self.dummies`` does not
        contain nans.

        Input:
        - Series with values
        """

        # Setup
        ohe = OneHotEncoder()

        # Run
        data = pd.Series(['a', 2, 'c'])
        ohe._fit(data)

        # Assert
        np.testing.assert_array_equal(ohe.dummies, ['a', 2, 'c'])

    def test__fit_dummies_nans(self):
        """Test the ``_fit`` method without nans.

        Check that ``self.dummies`` contain ``np.nan``.

        Input:
        - Series with values
        """

        # Setup
        ohe = OneHotEncoder()

        # Run
        data = pd.Series(['a', 2, 'c', None])
        ohe._fit(data)

        # Assert
        np.testing.assert_array_equal(ohe.dummies, ['a', 2, 'c', np.nan])

    def test__fit_no_nans(self):
        """Test the ``_fit`` method without nans.

        Check that the settings of the transformer
        are properly set based on the input. Encoding
        should be activated

        Input:
        - Series with values
        """

        # Setup
        ohe = OneHotEncoder()

        # Run
        data = pd.Series(['a', 'b', 'c'])
        ohe._fit(data)

        # Assert
        np.testing.assert_array_equal(ohe.dummies, ['a', 'b', 'c'])
        np.testing.assert_array_equal(ohe._uniques, ['a', 'b', 'c'])
        assert ohe._dummy_encoded
        assert not ohe._dummy_na

    def test__fit_no_nans_numeric(self):
        """Test the ``_fit`` method without nans.

        Check that the settings of the transformer
        are properly set based on the input. Encoding
        should be deactivated

        Input:
        - Series with values
        """

        # Setup
        ohe = OneHotEncoder()

        # Run
        data = pd.Series([1, 2, 3])
        ohe._fit(data)

        # Assert
        np.testing.assert_array_equal(ohe.dummies, [1, 2, 3])
        np.testing.assert_array_equal(ohe._uniques, [1, 2, 3])
        assert not ohe._dummy_encoded
        assert not ohe._dummy_na

    def test__fit_nans(self):
        """Test the ``_fit`` method with nans.

        Check that the settings of the transformer
        are properly set based on the input. Encoding
        and NA should be activated.

        Input:
        - Series with containing nan values
        """

        # Setup
        ohe = OneHotEncoder()

        # Run
        data = pd.Series(['a', 'b', None])
        ohe._fit(data)

        # Assert
        np.testing.assert_array_equal(ohe.dummies, ['a', 'b', np.nan])
        np.testing.assert_array_equal(ohe._uniques, ['a', 'b'])
        assert ohe._dummy_encoded
        assert ohe._dummy_na

    def test__fit_nans_numeric(self):
        """Test the ``_fit`` method with nans.

        Check that the settings of the transformer
        are properly set based on the input. Encoding
        should be deactivated and NA activated.

        Input:
        - Series with containing nan values
        """

        # Setup
        ohe = OneHotEncoder()

        # Run
        data = pd.Series([1, 2, np.nan])
        ohe._fit(data)

        # Assert
        np.testing.assert_array_equal(ohe.dummies, [1, 2, np.nan])
        np.testing.assert_array_equal(ohe._uniques, [1, 2])
        assert not ohe._dummy_encoded
        assert ohe._dummy_na

    def test__fit_single(self):
        # Setup
        ohe = OneHotEncoder()

        # Run
        data = pd.Series(['a', 'a', 'a'])
        ohe._fit(data)

        # Assert
        np.testing.assert_array_equal(ohe.dummies, ['a'])

    def test__transform_no_nan(self):
        """Test the ``_transform`` method without nans.

        The values passed to ``_transform`` should be
        returned in a one-hot encoding representation.

        Input:
        - Series with values
        Output:
        - one-hot encoding of the input
        """
        # Setup
        ohe = OneHotEncoder()
        data = pd.Series(['a', 'b', 'c'])
        ohe._uniques = ['a', 'b', 'c']
        ohe._num_dummies = 3

        # Run
        out = ohe._transform_helper(data)

        # Assert
        expected = np.array([
            [1, 0, 0],
            [0, 1, 0],
            [0, 0, 1]
        ])
        np.testing.assert_array_equal(out, expected)

    def test__transform_no_nan_categorical(self):
        """Test the ``_transform`` method without nans.

        The values passed to ``_transform`` should be
        returned in a one-hot encoding representation
        using the categorical branch.

        Input:
        - Series with categorical values
        Output:
        - one-hot encoding of the input
        """
        # Setup
        ohe = OneHotEncoder()
        data = pd.Series(['a', 'b', 'c'])
        ohe._uniques = ['a', 'b', 'c']
        ohe._indexer = [0, 1, 2]
        ohe._num_dummies = 3
        ohe._dummy_encoded = True

        # Run
        out = ohe._transform_helper(data)

        # Assert
        expected = np.array([
            [1, 0, 0],
            [0, 1, 0],
            [0, 0, 1]
        ])
        np.testing.assert_array_equal(out, expected)

    def test__transform_nans_encoded(self):
        """Test the ``_transform`` method with nans.

        The values passed to ``_transform`` should be
        returned in a one-hot encoding representation. Null
        values should be represented by the same encoding.

        Input:
        - Series with values containing nans
        Output:
        - one-hot encoding of the input
        """
        # Setup
        ohe = OneHotEncoder()
        data = pd.Series([np.nan, None, 'a', 'b'])
        ohe._uniques = ['a', 'b']
        ohe._dummy_na = True
        ohe._num_dummies = 2

        # Run
        out = ohe._transform_helper(data)

        # Assert
        expected = np.array([
            [0, 0, 1],
            [0, 0, 1],
            [1, 0, 0],
            [0, 1, 0]
        ])
        np.testing.assert_array_equal(out, expected)

    def test__transform_nans_categorical(self):
        """Test the ``_transform`` method with nans.

        The values passed to ``_transform`` should be
        returned in a one-hot encoding representation using
        the categorical branch. Null values should be
        represented by the same encoding.

        Input:
        - Series with categorical values containing nans
        Output:
        - one-hot encoding of the input
        """
        # Setup
        ohe = OneHotEncoder()
        data = pd.Series([np.nan, None, 'a', 'b'])
        ohe._uniques = ['a', 'b']
        ohe._indexer = [0, 1]
        ohe._dummy_na = True
        ohe._num_dummies = 2
        ohe._dummy_encoded = True

        # Run
        out = ohe._transform_helper(data)

        # Assert
        expected = np.array([
            [0, 0, 1],
            [0, 0, 1],
            [1, 0, 0],
            [0, 1, 0]
        ])
        np.testing.assert_array_equal(out, expected)

    def test__transform_single_column(self):
        """Test the ``_transform`` with one category.

        The values passed to ``_transform`` should be
        returned in a one-hot encoding representation
        where it should be a single column.

        Input:
        - Series with a single category
        Output:
        - one-hot encoding of the input
        """
        # Setup
        ohe = OneHotEncoder()
        data = pd.Series(['a', 'a', 'a'])
        ohe._uniques = ['a']
        ohe._num_dummies = 1

        # Run
        out = ohe._transform_helper(data)

        # Assert
        expected = np.array([
            [1],
            [1],
            [1]
        ])
        np.testing.assert_array_equal(out, expected)

    def test__transform_single_categorical(self):
        """Test the ``_transform`` with one category.

        The values passed to ``_transform`` should be
        returned in a one-hot encoding representation
        using the categorical branch where it should
        be a single column.

        Input:
        - Series with a single category
        Output:
        - one-hot encoding of the input
        """
        # Setup
        ohe = OneHotEncoder()
        data = pd.Series(['a', 'a', 'a'])
        ohe._uniques = ['a']
        ohe._indexer = [0]
        ohe._num_dummies = 1
        ohe._dummy_encoded = True

        # Run
        out = ohe._transform_helper(data)

        # Assert
        expected = np.array([
            [1],
            [1],
            [1]
        ])
        np.testing.assert_array_equal(out, expected)

    def test__transform_zeros(self):
        """Test the ``_transform`` with unknown category.

        The values passed to ``_transform`` should be
        returned in a one-hot encoding representation
        where it should be a column of zeros.

        Input:
        - Series with unknown values
        Output:
        - one-hot encoding of the input
        """
        # Setup
        ohe = OneHotEncoder()
        pd.Series(['a'])
        ohe._uniques = ['a']
        ohe._num_dummies = 1

        # Run
        out = ohe._transform_helper(pd.Series(['b', 'b', 'b']))

        # Assert
        expected = np.array([
            [0],
            [0],
            [0]
        ])
        np.testing.assert_array_equal(out, expected)

    def test__transform_zeros_categorical(self):
        """Test the ``_transform`` with unknown category.

        The values passed to ``_transform`` should be
        returned in a one-hot encoding representation
        using the categorical branch where it should
        be a column of zeros.

        Input:
        - Series with categorical and unknown values
        Output:
        - one-hot encoding of the input
        """
        # Setup
        ohe = OneHotEncoder()
        pd.Series(['a'])
        ohe._uniques = ['a']
        ohe._indexer = [0]
        ohe._num_dummies = 1
        ohe.dummy_encoded = True

        # Run
        out = ohe._transform_helper(pd.Series(['b', 'b', 'b']))

        # Assert
        expected = np.array([
            [0],
            [0],
            [0]
        ])
        np.testing.assert_array_equal(out, expected)

    def test__transform_unknown_nan(self):
        """Test the ``_transform`` with unknown and nans.

        This is an edge case for ``_transform`` where
        unknowns should be zeros and nans should be
        the last entry in the column.

        Input:
        - Series with unknown and nans
        Output:
        - one-hot encoding of the input
        """
        # Setup
        ohe = OneHotEncoder()
        pd.Series(['a'])
        ohe._uniques = ['a']
        ohe._dummy_na = True
        ohe._num_dummies = 1

        # Run
        out = ohe._transform_helper(pd.Series(['b', 'b', np.nan]))

        # Assert
        expected = np.array([
            [0, 0],
            [0, 0],
            [0, 1]
        ])
        np.testing.assert_array_equal(out, expected)

    def test__transform_no_nans(self):
        """Test the ``transform`` without nans.

        In this test ``transform`` should return an identity
        matrix representing each item in the input.

        Input:
        - Series with categorical values
        Output:
        - one-hot encoding of the input
        """
        # Setup
        ohe = OneHotEncoder()
        data = pd.Series(['a', 'b', 'c'])
        ohe._fit(data)

        # Run
        out = ohe._transform(data)

        # Assert
        expected = np.array([
            [1, 0, 0],
            [0, 1, 0],
            [0, 0, 1]
        ])
        np.testing.assert_array_equal(out, expected)

    def test__transform_nans(self):
        """Test the ``transform`` with nans.

        In this test ``transform`` should return an identity matrix
        representing each item in the input as well as nans.

        Input:
        - Series with categorical values and nans
        Output:
        - one-hot encoding of the input
        """
        # Setup
        ohe = OneHotEncoder()
        data = pd.Series(['a', 'b', None])
        ohe._fit(data)

        # Run
        out = ohe._transform(data)

        # Assert
        expected = np.array([
            [1, 0, 0],
            [0, 1, 0],
            [0, 0, 1]
        ])
        np.testing.assert_array_equal(out, expected)

    def test__transform_single_column_filled_with_ones(self):
        """Test the ``transform`` on a single category.

        In this test ``transform`` should return a column
        filled with ones.

        Input:
        - Series with a single categorical value
        Output:
        - one-hot encoding of the input
        """
        # Setup
        ohe = OneHotEncoder()
        data = pd.Series(['a', 'a', 'a'])
        ohe._fit(data)

        # Run
        out = ohe._transform(data)

        # Assert
        expected = np.array([
            [1],
            [1],
            [1]
        ])
        np.testing.assert_array_equal(out, expected)

    def test__transform_unknown(self):
        """Test the ``transform`` with unknown data.

        In this test ``transform`` should raise a warning due to the attempt
        of transforming data with previously unseen categories.

        Input:
        - Series with unknown categorical values
        Output:
        - one-hot encoding of the input, with the unseen category encoded as 0s
        """
        # Setup
        ohe = OneHotEncoder()
        fit_data = pd.Series([1, 2, 3, np.nan])
        ohe._fit(fit_data)

        # Run
        warning_msg = re.escape(
            'The data contains 1 new categories that were not '
            'seen in the original data (examples: {4.0}). Creating '
            'a vector of all 0s. If you want to model new categories, '
            'please fit the transformer again with the new data.'
        )
        with pytest.warns(UserWarning, match=warning_msg):
            transform_data = pd.Series([1, 2, np.nan, 4])
            out = ohe._transform(transform_data)

        # Assert
        expected = np.array([
            [1, 0, 0, 0],
            [0, 1, 0, 0],
            [0, 0, 0, 1],
            [0, 0, 0, 0]
        ])
        np.testing.assert_array_equal(out, expected)

    def test__transform_numeric(self):
        """Test the ``transform`` on numeric input.

        In this test ``transform`` should return a matrix
        representing each item in the input as one-hot encodings.

        Input:
        - Series with numeric input
        Output:
        - one-hot encoding of the input
        """
        # Setup
        ohe = OneHotEncoder()
        data = pd.Series([1, 2])
        ohe._fit(data)

        expected = np.array([
            [1, 0],
            [0, 1],
        ])

        # Run
        out = ohe._transform(data)

        # Assert
        assert not ohe._dummy_encoded
        np.testing.assert_array_equal(out, expected)

    def test__reverse_transform_no_nans(self):
        # Setup
        ohe = OneHotEncoder()
        data = pd.Series(['a', 'b', 'c'])
        ohe._fit(data)

        # Run
        transformed = np.array([
            [1, 0, 0],
            [0, 1, 0],
            [0, 0, 1]
        ])
        out = ohe._reverse_transform(transformed)

        # Assert
        expected = pd.Series(['a', 'b', 'c'])
        pd.testing.assert_series_equal(out, expected)

    def test__reverse_transform_nans(self):
        # Setup
        ohe = OneHotEncoder()
        data = pd.Series(['a', 'b', None])
        ohe._fit(data)

        # Run
        transformed = np.array([
            [1, 0, 0],
            [0, 1, 0],
            [0, 0, 1]
        ])
        out = ohe._reverse_transform(transformed)

        # Assert
        expected = pd.Series(['a', 'b', None])
        pd.testing.assert_series_equal(out, expected)

    def test__reverse_transform_single(self):
        # Setup
        ohe = OneHotEncoder()
        data = pd.Series(['a', 'a', 'a'])
        ohe._fit(data)

        # Run
        transformed = np.array([
            [1],
            [1],
            [1]
        ])
        out = ohe._reverse_transform(transformed)

        # Assert
        expected = pd.Series(['a', 'a', 'a'])
        pd.testing.assert_series_equal(out, expected)

    def test__reverse_transform_1d(self):
        # Setup
        ohe = OneHotEncoder()
        data = pd.Series(['a', 'a', 'a'])
        ohe._fit(data)

        # Run
        transformed = pd.Series([1, 1, 1])
        out = ohe._reverse_transform(transformed)

        # Assert
        expected = pd.Series(['a', 'a', 'a'])
        pd.testing.assert_series_equal(out, expected)


class TestLabelEncoder:

    def test___init__(self):
        """Passed arguments must be stored as attributes."""
        # Run
        transformer = LabelEncoder(add_noise='add_noise_value', order_by='alphabetical')

        # Asserts
        assert transformer.add_noise == 'add_noise_value'
        assert transformer.order_by == 'alphabetical'

    def test___init___bad_order_by(self):
        """Test that the ``__init__`` raises error if ``order_by`` is a bad value.

        Input:
            - ``order_by`` will be set to an unexpected string.

        Expected behavior:
            - An error should be raised.
        """
        # Run / Assert
        message = (
            "order_by must be one of the following values: None, 'numerical_value' or "
            "'alphabetical'"
        )
        with pytest.raises(Error, match=message):
            LabelEncoder(order_by='bad_value')

    def test__order_categories_alphabetical(self):
        """Test the ``_order_categories`` method when ``order_by`` is 'alphabetical'.

        Setup:
            - Set ``order_by`` to 'alphabetical'.

        Input:
            - numpy array of strings that are unordered.

        Output:
            - Same numpy array but with the strings alphabetically ordered.
        """
        # Setup
        transformer = LabelEncoder(order_by='alphabetical')
        arr = np.array(['one', 'two', 'three', 'four'])

        # Run
        ordered = transformer._order_categories(arr)

        # Assert
        np.testing.assert_array_equal(ordered, np.array(['four', 'one', 'three', 'two']))

    def test__order_categories_alphabetical_with_nans(self):
        """Test the ``_order_categories`` method when ``order_by`` is 'alphabetical'.

        Setup:
            - Set ``order_by`` to 'alphabetical'.

        Input:
            - numpy array of strings that are unordered and have nans.

        Output:
            - Same numpy array but with the strings alphabetically ordered and nan at the end.
        """
        # Setup
        transformer = LabelEncoder(order_by='alphabetical')
        arr = np.array(['one', 'two', 'three', np.nan, 'four'], dtype='object')

        # Run
        ordered = transformer._order_categories(arr)

        # Assert
        expected = np.array(['four', 'one', 'three', 'two', np.nan], dtype='object')
        pd.testing.assert_series_equal(pd.Series(ordered), pd.Series(expected))

    def test__order_categories_alphabetical_error(self):
        """Test the ``_order_categories`` method when ``order_by`` is 'alphabetical'.

        If ``order_by`` is 'alphabetical' but the data isn't a string, then an error should
        be raised.

        Setup:
            - Set ``order_by`` to 'alphabetical'.

        Input:
            - numpy array of strings that are unordered.

        Output:
            - Same numpy array but with the strings alphabetically ordered.
        """
        # Setup
        transformer = LabelEncoder(order_by='alphabetical')
        arr = np.array([1, 2, 3, 4])

        # Run / Assert
        message = "The data must be of type string if order_by is 'alphabetical'."
        with pytest.raises(Error, match=message):
            transformer._order_categories(arr)

    def test__order_categories_numerical(self):
        """Test the ``_order_categories`` method when ``order_by`` is 'numerical_value'.

        Setup:
            - Set ``order_by`` to 'numerical_value'.

        Input:
            - numpy array of numbers that are unordered.

        Output:
            - Same numpy array but with the numbers ordered.
        """
        # Setup
        transformer = LabelEncoder(order_by='numerical_value')
        arr = np.array([5, 3.11, 100, 67.8, np.nan, -2.5])

        # Run
        ordered = transformer._order_categories(arr)

        # Assert
        np.testing.assert_array_equal(ordered, np.array([-2.5, 3.11, 5, 67.8, 100, np.nan]))

    def test__order_categories_numerical_error(self):
        """Test the ``_order_categories`` method when ``order_by`` is 'numerical_value'.

        If the array is made up of strings that can't be converted to floats, and `order_by`
        is 'numerical_value', then we should raise an error.

        Setup:
            - Set ``order_by`` to 'numerical_value'.

        Input:
            - numpy array of strings.

        Expected behavior:
            - Error should be raised.
        """
        # Setup
        transformer = LabelEncoder(order_by='numerical_value')
        arr = np.array(['one', 'two', 'three', 'four'])

        # Run / Assert
        message = ("The data must be numerical if order_by is 'numerical_value'.")
        with pytest.raises(Error, match=message):
            transformer._order_categories(arr)

    def test__order_categories_numerical_different_dtype_error(self):
        """Test the ``_order_categories`` method when ``order_by`` is 'numerical_value'.

        If the array is made up of a dtype that is not numeric and can't be converted to a float,
        and `order_by` is 'numerical_value', then we should raise an error.

        Setup:
            - Set ``order_by`` to 'numerical_value'.

        Input:
            - numpy array of booleans.

        Expected behavior:
            - Error should be raised.
        """
        # Setup
        transformer = LabelEncoder(order_by='numerical_value')
        arr = np.array([True, False, False, True])

        # Run / Assert
        message = ("The data must be numerical if order_by is 'numerical_value'.")
        with pytest.raises(Error, match=message):
            transformer._order_categories(arr)

    def test__fit(self):
        """Test the ``_fit`` method.

        Validate that a unique integer representation for each category of the data is stored
        in the ``categories_to_values`` attribute, and the reverse is stored in the
        ``values_to_categories`` attribute .

        Setup:
            - create an instance of the ``LabelEncoder``.

        Input:
            - a pandas series.

        Side effects:
            - set the ``values_to_categories`` dictionary to the appropriate value.
            - set ``categories_to_values`` dictionary to the appropriate value.
        """
        # Setup
        data = pd.Series([1, 2, 3, 2, 1])
        transformer = LabelEncoder()

        # Run
        transformer._fit(data)

        # Assert
        assert transformer.values_to_categories == {0: 1, 1: 2, 2: 3}
        assert transformer.categories_to_values == {1: 0, 2: 1, 3: 2}

    def test__transform(self):
        """Test the ``_transform`` method.

        Validate that each category of the passed data is replaced with its corresponding
        integer value.

        Setup:
            - create an instance of the ``LabelEncoder``, where ``categories_to_values``
            and ``values_to_categories`` are set to dictionaries.

        Input:
            - a pandas series.

        Output:
            - a numpy array containing the transformed data.
        """
        # Setup
        data = pd.Series([1, 2, 3, 4])
        transformer = LabelEncoder()
        transformer.categories_to_values = {1: 0, 2: 1, 3: 2}
        transformer.values_to_categories = {0: 1, 1: 2, 2: 3}

        # Run
        warning_msg = re.escape(
            'The data contains 1 new categories that were not '
            'seen in the original data (examples: {4}). Assigning '
            'them random values. If you want to model new categories, '
            'please fit the transformer again with the new data.'
        )
        with pytest.warns(UserWarning, match=warning_msg):
            transformed = transformer._transform(data)

        # Assert
        expected = pd.Series([0., 1., 2.])
        pd.testing.assert_series_equal(transformed[:-1], expected)

        assert 0 <= transformed[3] <= 2

    def test__transform_add_noise(self):
        """Test the ``_transform`` method with ``add_noise``.

        Validate that the method correctly transforms the categories when ``add_noise`` is True.

        Setup:
            - create an instance of the ``LabelEncoder``, where ``categories_to_values``
            and ``values_to_categories`` are set to dictionaries.
            - set ``add_noise`` to True.

        Input:
            - a pandas series.

        Output:
            - a numpy array containing the transformed data.
        """
        # Setup
        data = pd.Series([1, 2, 3, 4])
        transformer = LabelEncoder(add_noise=True)
        transformer.categories_to_values = {1: 0, 2: 1, 3: 2}
        transformer.values_to_categories = {0: 1, 1: 2, 2: 3}

        # Run
        transformed = transformer._transform(data)

        # Assert
        assert 0 <= transformed[0] < 1
        assert 1 <= transformed[1] < 2
        assert 2 <= transformed[2] < 3
        assert 0 <= transformed[3] < 3

    def test__transform_unseen_categories(self):
        """Test the ``_transform`` method with multiple unseen categories.

        Validate that each category of the passed data is replaced with its corresponding
        integer value.

        Setup:
            - create an instance of the ``LabelEncoder``, where ``categories_to_values``
            and ``values_to_categories`` are set to dictionaries.

        Input:
            - a pandas series.

        Output:
            - a numpy array containing the transformed data.
        """
        # Setup
        fit_data = pd.Series(['a', 2, True])
        transformer = LabelEncoder()
        transformer.categories_to_values = {'a': 0, 2: 1, True: 2}
        transformer.values_to_categories = {0: 'a', 1: 2, 2: True}

        # Run
        with pytest.warns(UserWarning):
            transform_data = pd.Series(['a', 2, True, np.nan, np.nan, np.nan, 'b', False, 3])
            transformed = transformer._transform(transform_data)

        # Assert
        expected = pd.Series([0., 1., 2.])
        pd.testing.assert_series_equal(transformed[:3], expected)

        assert all([0 <= value < len(fit_data) for value in transformed[3:]])

    def test__reverse_transform_clips_values(self):
        """Test the ``_reverse_transform`` method with values not in map.

        If a value that is not in ``values_to_categories`` is passed
        to ``reverse_transform``, then the value should be clipped to
        the range of the dict's keys.

        Input:
        - array with values outside of dict
        Output:
        - categories corresponding to closest key in the dict
        """
        # Setup
        transformer = LabelEncoder()
        transformer.values_to_categories = {0: 'a', 1: 'b', 2: 'c'}
        data = pd.Series([0, 1, 10])

        # Run
        out = transformer._reverse_transform(data)

        # Assert
        pd.testing.assert_series_equal(out, pd.Series(['a', 'b', 'c']))

    def test__reverse_transform_add_noise(self):
        """Test the ``_reverse_transform`` method with ``add_noise``.

        Test that the method correctly reverse transforms the data
        when ``add_noise`` is set to True.

        Input:
            - pd.Series
        Output:
            - corresponding categories
        """
        # Setup
        transformer = LabelEncoder(add_noise=True)
        transformer.values_to_categories = {0: 'a', 1: 'b', 2: 'c'}
        data = pd.Series([0.5, 1.0, 10.9])

        # Run
        out = transformer._reverse_transform(data)

        # Assert
        pd.testing.assert_series_equal(out, pd.Series(['a', 'b', 'c']))


class TestCustomLabelEncoder:

    def test___init__(self):
        """The the ``__init__`` method.

        Passed arguments must be stored as attributes.
        """
        # Run
        transformer = CustomLabelEncoder(order=['b', 'c', 'a', None], add_noise='add_noise_value')

        # Asserts
        assert transformer.add_noise == 'add_noise_value'
        pd.testing.assert_series_equal(transformer.order, pd.Series(['b', 'c', 'a', np.nan]))

    def test__fit(self):
        """Test the ``_fit`` method.

        Validate that a unique integer representation for each category of the data is stored
        in the ``categories_to_values`` attribute, and the reverse is stored in the
        ``values_to_categories`` attribute. The order should match the ``self.order`` indices.

        Setup:
            - create an instance of the ``CustomLabelEncoder``.

        Input:
            - a pandas series.

        Side effects:
            - set the ``values_to_categories`` dictionary to the appropriate value.
            - set ``categories_to_values`` dictionary to the appropriate value.
        """
        # Setup
        data = pd.Series([1, 2, 3, 2, np.nan, 1])
        transformer = CustomLabelEncoder(order=[2, 3, np.nan, 1])

        # Run
        transformer._fit(data)

        # Assert
        expected_values_to_categories = {0: 2, 1: 3, 2: np.nan, 3: 1}
        expected_categories_to_values = {2: 0, 3: 1, 1: 3, np.nan: 2}
        for key, value in transformer.values_to_categories.items():
            assert value == expected_values_to_categories[key] or pd.isna(value)

        for key, value in transformer.categories_to_values.items():
            assert value == expected_categories_to_values.get(key) or pd.isna(key)

    def test__fit_error(self):
        """Test the ``_fit`` method checks that data is in ``self.order``.

        If the data being fit is not in ``self.order`` an error should be raised.

        Setup:
            - create an instance of the ``CustomLabelEncoder``.

        Input:
            - a pandas series.

        Side effects:
            - set the ``values_to_categories`` dictionary to the appropriate value.
            - set ``categories_to_values`` dictionary to the appropriate value.
        """
        # Setup
        data = pd.Series([1, 2, 3, 2, 1, 4])
        transformer = CustomLabelEncoder(order=[2, 1])

        # Run / Assert
        message = re.escape(
            "Unknown categories '[3, 4]'. All possible categories must be defined in the "
            "'order' parameter."
        )
        with pytest.raises(Error, match=message):
            transformer._fit(data)



import os
from collections import defaultdict

import numpy as np
import pandas as pd
from sklearn.linear_model import LinearRegression
from sklearn.model_selection import cross_val_score

from rdt import HyperTransformer
from rdt.transformers import CustomLabelEncoder, FloatFormatter, get_transformers_by_type
from tests.quality.utils import download_single_table

R2_THRESHOLD = 0.2
TEST_THRESHOLD = 0.3
MAX_SIZE = 5000000
SDTYPES_TO_SKIP = {'numerical', 'float', 'integer', 'id', None}

TYPE_TO_DTYPE = {
    'numerical': ['number'],
    'float': ['float'],
    'int': ['int'],
    'categorical': ['object', 'category'],
    'datetime': ['datetime'],
    'boolean': ['bool']
}
TRANSFORMERS_TO_SKIP = [CustomLabelEncoder]


def format_array(array):
    if not isinstance(array, np.ndarray):
        array = array.to_numpy()

    if len(array.shape) == 1:
        array = array.reshape(-1, 1)

    return array


def get_regression_score(features, target):
    model = LinearRegression()
    scores = cross_val_score(model, features, target)
    return np.mean(scores)


def find_columns(data, sdtype, metadata=None):
    if metadata:
        return {
            column
            for column in metadata['fields']
            if metadata['fields'][column]['type'] == sdtype
        }

    columns = set()
    dtypes = TYPE_TO_DTYPE.get(sdtype, sdtype)
    for dtype in dtypes:
        selected = data.select_dtypes(dtype)
        columns.update(set(selected.columns))

    return columns


def get_transformer_regression_scores(data, sdtype, dataset_name, transformers, metadata=None):
    """Returns regression scores for a list of transformers.

    Args:
        data (pandas.DataFrame):
            The dataset containing columns to predict and train with.
        sdtype (string):
            The sdtype of the transformer.
        dataset_name (string):
            The name of the dataset.
        transformers (list):
            List of transformer instances.
        metadata (dict):
            Dictionary containing metadata for the table.

    Returns:
        pandas.DataFrame containing the score for each column predicted
        in the dataset. To get the scores, a regression model is trained.
        The features used are the output of transforming all the columns
        of the sdtype using a transformer in the transformers list.
    """
    columns_to_predict = find_columns(data, 'numerical')
    columns_to_transform = find_columns(data, sdtype, metadata)
    scores = pd.DataFrame()
    features = data[columns_to_transform]

    for column in columns_to_predict:
        target = data[column].to_frame()
        numerical_transformer = FloatFormatter(model_missing_values=False)
        target = numerical_transformer.fit_transform(target, column)
        target = format_array(target)
        nans = np.isnan(target)[:, 0]
        target = target[~nans]
        for transformer in transformers:
            ht = HyperTransformer()
            ht.detect_initial_config(features)
            ht.update_transformers_by_sdtype(sdtype=sdtype, transformer=transformer())
            ht.fit(features)
            transformed_features = ht.transform(features).to_numpy()
            transformed_features = transformed_features[~nans]
            score = get_regression_score(transformed_features, target)
            row = pd.Series({
                'transformer_name': transformer.__name__,
                'dataset_name': dataset_name,
                'column': column,
                'score': score
            })
            scores = scores.append(row, ignore_index=True)

    return scores


def get_test_cases(sdtypes):
    test_cases = []
    path = os.path.join(os.path.dirname(__file__), 'dataset_info.csv')
    datasets = pd.read_csv(path)
    for _, row in datasets.iterrows():
        if row['table_size'] < MAX_SIZE and row['modality'] == 'single-table':
            table_types = eval(row['table_types'])
            table_types_to_test = sdtypes.intersection(table_types)
            if len(table_types_to_test) > 0:
                test_cases.append((row['name'], row['table_name'], table_types_to_test))

    return test_cases


def get_regression_scores(test_cases, transformers_by_type):
    """Create table of all regression scores for test cases.

    Args:
        test_cases (list):
            List of test cases. Each test case is a tuple containing
            the dataset name, the name of the table to use from the
            dataset, and the sdtypes to test against for that table.
        transformers_by_type (dict):
            Dict mapping sdtype to list of transformers that have that
            type as their input sdtype.

    Returns:
        DataFrame where each row has a dataset name, transformer name,
        column name and regression score. The regression score is the
        coefficient of determination for the transformer predicting the column.
    """
    all_scores = defaultdict(pd.DataFrame)
    for dataset_name, table_name, sdtypes in test_cases:
        (data, metadata) = download_single_table(dataset_name, table_name)
        for sdtype in sdtypes:
            transformers = transformers_by_type[sdtype]
            regression_scores = get_transformer_regression_scores(
                data, sdtype, dataset_name, transformers, metadata)
            all_scores[sdtype] = all_scores[sdtype].append(
                regression_scores, ignore_index=True)

    return all_scores


def get_results_table(regression_scores):
    """Create a table of results for each transformer on each dataset.

    Args:
        regression_scores (dict):
            Dict mapping sdtypes to a DataFrame where each row has
            a table name, column name, transformer name and coefficient
            of determination for that transformer predicting that column.

    Returns:
        A DataFrame where each row has a transformer name, dataset name,
        average score for the dataset and a score comparing the transformer's
        average score for the dataset to the average of the average score for
        the dataset across all transformers of the same sdtype.
    """
    results = pd.DataFrame()
    for _, scores in regression_scores.items():
        table_column_groups = scores.groupby(['dataset_name', 'column'])
        valid = []
        for _, frame in table_column_groups:
            if frame['score'].mean() >= R2_THRESHOLD:
                valid.extend(frame.index)

        valid_scores = scores.loc[valid]
        transformer_dataset_groups = valid_scores.groupby(['dataset_name', 'transformer_name'])
        for (dataset_name, transformer_name), frame in transformer_dataset_groups:
            transformer_average = frame['score'].mean()
            dataset_rows = (valid_scores['dataset_name'] == dataset_name)
            transformer_rows = (valid_scores['transformer_name'] != transformer_name)
            data_without_transformer = valid_scores.loc[
                dataset_rows & transformer_rows
            ]
            average_without_transformer = data_without_transformer['score'].mean()

            row = pd.Series({
                'transformer_name': transformer_name,
                'dataset_name': dataset_name,
                'score': transformer_average,
                'score_relative_to_average': transformer_average / average_without_transformer
            })
            results = results.append(row, ignore_index=True)

    return results


def test_quality(subtests):
    """Run all the quality test cases.

    This test has multiple steps.
        1. It creates a list of test cases. Each test case has a dataset
        and a set of sdtypes to test for the dataset.
        2. A dictionary is created mapping sdtypes to a DataFrame
        containing the regression scores obtained from running the
        transformers of that sdtype against the datasets in the test cases.
        Each row in the DataFrame has the transformer name, dataset name,
        column name and score. The scores are computed as follows:
            - For every transformer of the sdtype, transform all the
            columns of that sdtype.
            - For every numerical column in the dataset, the transformed
            columns are used as features to train a regression model.
            - The score is the coefficient of determination obtained from
            that model trying to predict the target column.
        3. Once the scores are gathered, a results table is created. Each row has
        a transformer name, dataset name, average score for the dataset and a score
        comparing the transformer's average score for the dataset to the average
        of the average score for the dataset across all transformers of the same
        sdtype.
        4. For every unique transformer in the results, a test is run to check
        that the transformer's score for each table is either higher than the
        threshold, or the comparitive score is higher than the threshold.
    """
    transformers_by_type = get_transformers_by_type()
    for transformer_list in transformers_by_type.values():
        for transformer in transformer_list:
            if transformer in TRANSFORMERS_TO_SKIP:
                transformer_list.remove(transformer)

    sdtypes_to_test = {
        sdtype
        for sdtype in transformers_by_type.keys()
        if sdtype not in SDTYPES_TO_SKIP
    }
    test_cases = get_test_cases(sdtypes_to_test)
    all_regression_scores = get_regression_scores(test_cases, transformers_by_type)
    results = get_results_table(all_regression_scores)

    for transformer, frame in results.groupby('transformer_name'):
        with subtests.test(
                msg=f'Testing transformer {transformer}',
                transformer=transformer):
            relative_scores = frame['score_relative_to_average']
            assert all((relative_scores > TEST_THRESHOLD) | (frame['score'] > TEST_THRESHOLD))



"""Test Personal Identifiable Information Transformer using Faker."""

from unittest.mock import Mock, call, patch

import numpy as np
import pandas as pd
import pytest

from rdt.errors import Error
from rdt.transformers.null import NullTransformer
from rdt.transformers.pii.anonymizer import AnonymizedFaker


class TestAnonymizedFaker:
    """Test class for ``AnonymizedFaker``."""

    @patch('rdt.transformers.pii.anonymizer.faker')
    @patch('rdt.transformers.pii.anonymizer.getattr')
    def test_check_provider_function_baseprovider(self, mock_getattr, mock_faker):
        """Test that ``getattr`` is being called with ``BaseProvider`` and ``function_name``.

        Mock:
            - Mock the ``getattr`` from Python to ensure that is being called with the input.
            - Mock faker and ensure that ``getattr`` is being called with ``faker.providers``.
        """
        # Setup
        mock_getattr.side_effect = ['module', 'provider', None]

        # Run
        AnonymizedFaker.check_provider_function('BaseProvider', 'function_name')

        # Assert
        assert mock_getattr.call_args_list[0] == call(mock_faker.providers, 'BaseProvider')
        assert mock_getattr.call_args_list[1] == call('module', 'function_name')

    @patch('rdt.transformers.pii.anonymizer.faker')
    @patch('rdt.transformers.pii.anonymizer.getattr')
    def test_check_provider_function_other_providers(self, mock_getattr, mock_faker):
        """Test that ``getattr`` is being called with ``provider_name`` and ``function_name``.

        Mock:
            - Mock the ``getattr`` from Python to ensure that is being called with the input.
            - Mock faker and ensure that ``getattr`` is being called with ``faker.providers``.
        """
        # Setup
        mock_getattr.side_effect = ['module', 'provider_class', None]

        # Run
        AnonymizedFaker.check_provider_function('provider_name', 'function_name')

        # Assert
        assert mock_getattr.call_args_list[0] == call(mock_faker.providers, 'provider_name')
        assert mock_getattr.call_args_list[1] == call('module', 'Provider')
        assert mock_getattr.call_args_list[2] == call('provider_class', 'function_name')

    def test_check_provider_function_raise_attribute_error(self):
        """Test that ``check_provider_function`` raises an ``AttributeError``.

        Test when the provider class is not in the ``faker.providers`` an ``AttributeError`` is
        being raised or when the ``function`` is not within the ``provider``.
        """
        # Setup
        expected_message = (
            "The 'TestProvider' module does not contain a function named "
            "'TestFunction'.\nRefer to the Faker docs to find the correct function: "
            'https://faker.readthedocs.io/en/master/providers.html'
        )

        # Run
        with pytest.raises(Error, match=expected_message):
            AnonymizedFaker.check_provider_function('TestProvider', 'TestFunction')

    def test__function(self):
        """Test that `_function`.

        The method `_function` should return a call from the `instance.faker.provider.<function>`.

        Mock:
            - Instance of 'AnonymizedFaker'.
            - Faker instance.
            - A function for the faker instance.

        Output:
            - Return value of mocked function.

        Side Effects:
            - The returned function, when called, has to call the `faker.<function_name>` function
              with the provided kwargs.
        """
        # setup
        instance = Mock()
        function = Mock()
        function.return_value = 1
        instance.faker.number = function
        instance.function_name = 'number'
        instance.function_kwargs = {'type': 'int'}

        # Run
        result = AnonymizedFaker._function(instance)

        # Assert
        function.assert_called_once_with(type='int')
        assert result == 1

    @patch('rdt.transformers.pii.anonymizer.importlib')
    @patch('rdt.transformers.pii.anonymizer.warnings')
    def test__check_locales(self, mock_warnings, mock_importlib):
        """Test that check locales warns the user if the spec was not found.

        Mock:
            - Mock importlib with side effects to return one `None` and one value.
            - Mock the warnings.
        Side Effect:
            - mock_warnings has been called once with the expected message.
        """
        # Setup
        instance = Mock()
        instance.provider_name = 'credit_card'
        instance.function_name = 'credit_card_full'
        instance.locales = ['es_ES', 'en_US']
        mock_importlib.util.find_spec.side_effect = [None, 'en_US']

        # Run
        AnonymizedFaker._check_locales(instance)

        # Assert
        expected_message = (
            "Locales ['es_ES'] do not support provider 'credit_card' "
            "and function 'credit_card_full'.\nIn place of these locales, 'en_US' will "
            'be used instead. Please refer to the localized provider docs for more '
            'information: https://faker.readthedocs.io/en/master/locales.html'
        )
        mock_warnings.warn.assert_called_once_with(expected_message)

    @patch('rdt.transformers.pii.anonymizer.faker')
    @patch('rdt.transformers.pii.anonymizer.AnonymizedFaker.check_provider_function')
    def test___init__default(self, mock_check_provider_function, mock_faker):
        """Test the default instantiation of the transformer.

        Test that by default the transformer is being instantiated with ``BaseProvider`` and
        ``lexify``.

        Mock:
            - ``check_provider_function`` mock and assert that this is being called with
              ``BaseProvider`` and ``lexify``.
            - ``faker`` mock to ensure that is being called with ``None`` as locales
              and that the ``lexify`` from the instance is being assigned to the ``_function``.

        Side effects:
            - the ``instance.provider_name`` is ``BaseProvider``.
            - the ``instance.function_name`` is ``lexify``.
            - the ``instance.locales`` is ``None``.
            - the ``instance.function_kwargs`` is an empty ``dict``.
            - the ``instance.missing_value_replacement`` is ``None``.
            - the ``instance.model_missing_values`` is ``False``
            - ``check_provider_function`` has been called once with ``BaseProvider`` and
              ``lexify``.
            - the ``instance._function`` is ``instance.faker.lexify``.
        """
        # Run
        instance = AnonymizedFaker()

        # Assert
        mock_check_provider_function.assert_called_once_with('BaseProvider', 'lexify')
        assert instance.provider_name == 'BaseProvider'
        assert instance.function_name == 'lexify'
        assert instance.function_kwargs == {}
        assert instance.missing_value_replacement is None
        assert not instance.model_missing_values
        assert instance.locales is None
        assert mock_faker.Faker.called_once_with(None)

    @patch('rdt.transformers.pii.anonymizer.faker')
    @patch('rdt.transformers.pii.anonymizer.AnonymizedFaker.check_provider_function')
    def test___init__custom(self, mock_check_provider_function, mock_faker):
        """Test the instantiation of the transformer with custom parameters.

        Test that the transformer can be instantiated with a custom provider and function, and
        this is being assigned properly.

        Mock:
            - ``check_provider_function`` mock and assert that this is being called with
              ``CreditCard`` and ``credit_card_full``.
            - ``faker`` mock to ensure that is being called with ``['en_US', 'fr_FR']`` as locales
              and that the ``credit_card_full`` from the instance is being assigned to the
              ``_function``.

        Side effects:
            - the ``instance.provider_name`` is ``CreditCard``.
            - the ``instance.function_name`` is ``credit_card_full``.
            - the ``instance.locales`` is ``['en_US', 'fr_FR']``.
            - the ``instance.function_kwargs`` is ``{'type': 'visa'}``.
            - the ``instance.missing_value_replacement`` is ``None``.
            - the ``instance.model_missing_values`` is ``False``
            - ``check_provider_function`` has been called once with ``CreditCard`` and
              ``credit_card_full``.
            - the ``instance._function`` is ``instance.faker.credit_card_full``.
        """
        # Run
        instance = AnonymizedFaker(
            provider_name='credit_card',
            function_name='credit_card_full',
            function_kwargs={
                'type': 'visa'
            },
            locales=['en_US', 'fr_FR']
        )

        # Assert
        mock_check_provider_function.assert_called_once_with('credit_card', 'credit_card_full')
        assert instance.provider_name == 'credit_card'
        assert instance.function_name == 'credit_card_full'
        assert instance.function_kwargs == {'type': 'visa'}
        assert instance.missing_value_replacement is None
        assert not instance.model_missing_values
        assert instance.locales == ['en_US', 'fr_FR']
        assert mock_faker.Faker.called_once_with(['en_US', 'fr_FR'])

    def test___init__no_function_name(self):
        """Test the instantiation of the transformer with custom parameters.

        Test that the transformer raises an error when no function name is provided.

        Raises:
            - Error.
        """
        # Run / Assert
        expected_message = (
            'Please specify the function name to use from the '
            "'credit_card' provider."
        )
        with pytest.raises(Error, match=expected_message):
            AnonymizedFaker(provider_name='credit_card', locales=['en_US', 'fr_FR'])

    def test_get_output_sdtypes(self):
        """Test the ``get_output_sdtypes``.

        Setup:
            - initialize a ``AnonymizedFaker`` transformer with default values.

        Output:
            - the ``output_sdtypes`` returns an empty dictionary.
        """
        # Setup
        transformer = AnonymizedFaker()
        transformer.column_prefix = 'a#b'

        # Run
        output = transformer.get_output_sdtypes()

        # Assert
        expected = {}
        assert output == expected

    def test_get_output_sdtypes_model_missing_values(self):
        """Test the ``get_output_sdtypes`` method when a null column is created.

        Setup:
            - initialize a ``AnonymizedFaker`` transformer which:
                - sets ``self.null_transformer`` to a ``NullTransformer`` where
                ``self.model_missing_values`` is True.
                - sets ``self.column_prefix`` to a string.

        Output:
            - An ``output_sdtypes`` dictionary is being returned with the ``self.column_prefix``
              added to the beginning of the keys.
        """
        # Setup
        transformer = AnonymizedFaker()
        transformer.null_transformer = NullTransformer(missing_value_replacement='fill')
        transformer.null_transformer._model_missing_values = True
        transformer.column_prefix = 'a#b'

        # Run
        output = transformer.get_output_sdtypes()

        # Assert
        expected = {
            'a#b.is_null': 'float'
        }
        assert output == expected

    @patch('rdt.transformers.pii.anonymizer.NullTransformer')
    def test__fit(self, mock_null_transformer):
        """Test the ``_fit`` method.

        Validate that the ``_fit`` method uses the ``NullTransformer`` to parse the data
        and learn the length of it.

        Setup:
            - Initialize a ``AnonymizedFaker`` transformer.
            - Mock the ``NullTransformer``.

        Input:
            - ``pd.Series`` containing 3 strings.

        Side Effects:
            - ``NullTransformer`` instance has been created with ``model_missing_values`` as
              ``False`` and ``missing_value_replacement`` as ``None``.
            - ``ǸullTransformer`` instance method ``fit`` has been called with the input data.
            - ``instance.data_length`` equals to the length of the input data.
        """
        # Setup
        transformer = AnonymizedFaker()

        columns_data = pd.Series(['1', '2', '3'])

        # Run
        transformer._fit(columns_data)

        # Assert
        mock_null_transformer.assert_called_once_with(None, False)
        mock_null_transformer.return_value.fit.called_once_with(columns_data)
        assert transformer.data_length == 3

    def test__transform(self):
        """Test the ``_transform`` method.

        Validate that the ``_transform`` method returns ``None`` when the ``NullTransformer``
        does not model the missing values.

        Setup:
            - Initialize a ``AnonymizedFaker`` transformer.

        Input:
            - ``pd.Series`` with three values.

        Output:
            - ``None``.
        """
        # Setup
        columns_data = pd.Series([1, 2, 3])
        instance = AnonymizedFaker()

        # Run
        result = instance._transform(columns_data)

        # Assert
        assert result is None

    def test__transform_model_missing_values(self):
        """Test the ``_transform`` method.

        Validate that the ``_transform`` method uses the ``NullTransformer`` instance to
        transform the data.

        Setup:
            - Initialize a ``AnonymizedFaker`` transformer.
            - Mock the ``null_transformer`` of the instance.
            - Mock the return value of the ``null_transformer.transform``.

        Input:
            - ``pd.Series`` with three values.

        Output:
            - The second dimension of the mocked return value of the
              ``null_transformer.transform``.
        """
        # Setup
        columns_data = pd.Series([1, 2, 3])
        instance = AnonymizedFaker()
        instance.null_transformer = Mock()

        instance.null_transformer.transform.return_value = np.array([
            [4, 0],
            [5, 1],
            [6, 0],
        ])

        # Run
        result = instance._transform(columns_data)

        # Assert
        instance.null_transformer.transform.assert_called_once_with(columns_data)
        np.testing.assert_array_equal(result, np.array([0, 1, 0]))

    def test__reverse_transform(self):
        """Test the ``_reverse_transform`` method.

        Validate that the ``_reverse_transform`` method calls the ``instance._function`` with
        the ``instance.function_kwargs`` the ``instance.data_length`` amount of times.

        Setup:
            - Initialize a ``AnonymizedFaker`` transformer.
            - Mock the ``null_transformer`` of the instance.
            - Mock the return value of the ``null_transformer.reverse_transform``.

        Input:
            - ``pd.Series`` with three values.

        Output:
            - the output of ``null_transformer.reverse_transform``.
        """
        # Setup
        instance = AnonymizedFaker()
        instance.null_transformer = Mock()
        instance.null_transformer.models_missing_values.return_value = False
        instance.data_length = 3
        function = Mock()
        function.side_effect = ['a', 'b', 'c']

        instance._function = function
        instance.null_transformer.reverse_transform.return_value = np.array(['a', 'b', 'c'])

        # Run
        result = instance._reverse_transform(None)

        # Assert
        expected_null_call = np.array(['a', 'b', 'c'])
        null_call = instance.null_transformer.reverse_transform.call_args_list[0][0][0]
        np.testing.assert_array_equal(null_call, expected_null_call)
        assert function.call_args_list == [call(), call(), call()]
        np.testing.assert_array_equal(result, np.array(['a', 'b', 'c']))

    def test__reverse_transform_models_missing_values(self):
        """Test the ``_reverse_transform`` method.

        Validate that the ``_reverse_transform`` method uses the ``instance._function``
        to generate values within the range of the ``instance.data_length``.

        Setup:
            - Mock the instance of ``AnonymizedFaker``.
            - Mock the ``instance.null_transformer.reverse_transform`` return value.

        Input:
            - ``pd.DataFrame`` with a column ``is_null`` that contains numeric values.

        Output:
            - The mocked return value of the ``null_transformer.reverse_transform``.

        Side Effects:
            - The ``instance._function`` has been called ``instance.data_length`` with
              the ``function_args`` as keyword args.
        """
        # Setup
        columns_data = pd.DataFrame({
            'is_null': [0, 1, 0]
        })
        instance = Mock()
        instance.null_transformer.reverse_transform.return_value = pd.Series([
            'a',
            np.nan,
            'c',
        ])
        instance.get_output_columns.return_value = ['is_null']
        function = Mock()
        function.side_effect = [1, 2, 3]
        instance._function = function
        instance.data_length = 3
        instance.function_kwargs = {
            'type': 'a'
        }

        # Run
        output = AnonymizedFaker._reverse_transform(instance, columns_data)

        # Assert
        expected_output = pd.Series([
            'a',
            np.nan,
            'c'
        ])
        pd.testing.assert_series_equal(expected_output, output)

        expected_call = np.array([
            [1, 0],
            [2, 1],
            [3, 0]
        ])
        called_arg = instance.null_transformer.reverse_transform.call_args[0][0]
        np.testing.assert_array_equal(expected_call, called_arg)

        expected_function_calls = [call(), call(), call()]
        assert function.call_args_list == expected_function_calls

    def test___repr__default(self):
        """Test the ``__repr__`` method.

        With the default parameters should return only the ``function_name='lexify'`` as an
        starting argument for the ``AnonymizedFaker``.
        """
        # Setup
        instance = AnonymizedFaker()

        # Run
        res = repr(instance)

        # Assert
        expected_res = "AnonymizedFaker(function_name='lexify')"
        assert res == expected_res

    def test___repr__custom_provider(self):
        """Test the ``__repr__`` method.

        With the custom args the ``repr`` of the class should return all the non default
        arguments.
        """
        # Setup
        instance = AnonymizedFaker('credit_card', 'credit_card_full', model_missing_values=True)

        # Run
        res = repr(instance)

        # Assert
        expected_res = (
            "AnonymizedFaker(provider_name='credit_card', function_name='credit_card_full', "
            'model_missing_values=True)'
        )

        assert res == expected_res



"""Basic Integration tests for the BaseTransformer."""

import numpy as np
import pandas as pd

from rdt.transformers.base import BaseTransformer


def test_dummy_transformer_series_output():
    """Test a transformer that outputs a Series.

    This validates that a Transformer that is implemented to
    produce a Series object as the output of the `_transform`
    method works as expected.

    Setup:
        - A DummyTransformer that implements a _transform
          that produces a Series and a reverse_transform
          that takes a Series as input.

    Input:
        - A DataFrame with a single column of boolean values

    Expected behavior:
        - The data should be transformed into a DataFrame that contains
          a single float column made of 0.0s and 1.0s.
        - The transformed data should be able to reversed to
          re-produce the input data.
    """
    # Setup
    class DummyTransformer(BaseTransformer):

        INPUT_SDTYPE = 'boolean'
        OUTPUT_SDTYPES = {
            'value': 'float'
        }

        def _fit(self, data):
            pass

        def _transform(self, data):
            return data.astype(float)

        def _reverse_transform(self, data):
            return data.round() != 0

    # Run
    data = pd.DataFrame({
        'bool': [True, False, True, False]
    })

    transformer = DummyTransformer()
    transformed = transformer.fit_transform(data, 'bool')

    reverse = transformer.reverse_transform(transformed)

    # Assert
    expected_transform = pd.DataFrame({
        'bool.value': [1., 0., 1., 0.]
    })
    pd.testing.assert_frame_equal(expected_transform, transformed)
    pd.testing.assert_frame_equal(reverse, data)


def test_dummy_transformer_dataframe_output():
    """Test a transformer that outputs a DataFrame.

    This validates that a Transformer that is implemented to
    produce a DataFrame object as the output of the `_transform`
    method works as expected.

    Setup:
        - A DummyTransformer that implements a _transform
          that produces a DataFrame and a reverse_transform
          that takes a DataFrame as input.

    Input:
        - A DataFrame with a single column of boolean values

    Expected behavior:
        - The data should be transformed into a DataFrame that contains
          a float column made of 0.0s, 1.0s and -1.0s, and another float
          column that contains 0.0s and 1.0s indicating which values were
          null in the input.
        - The transformed data should be able to reversed to
          re-produce the input data.
    """
    # Setup
    class DummyTransformer(BaseTransformer):

        INPUT_SDTYPE = 'boolean'
        OUTPUT_SDTYPES = {
            'value': 'float',
            'null': 'float'
        }

        def _fit(self, data):
            pass

        def _transform(self, data):
            out = pd.DataFrame(dict(zip(
                self.output_columns,
                [
                    data.astype(float).fillna(-1),
                    data.isna().astype(float)
                ]
            )))

            return out

        def _reverse_transform(self, data):
            output = data[self.output_columns[0]]
            output = output.round().astype(bool).astype(object)
            output.iloc[data[self.output_columns[1]] == 1] = np.nan

            return output

    # Run
    data = pd.DataFrame({
        'bool': [True, False, True, np.nan]
    })

    transformer = DummyTransformer()
    transformed = transformer.fit_transform(data, 'bool')

    reverse = transformer.reverse_transform(transformed)

    # Assert
    expected_transform = pd.DataFrame({
        'bool.value': [1., 0., 1., -1.],
        'bool.null': [0., 0., 0., 1.]
    })
    pd.testing.assert_frame_equal(expected_transform, transformed)
    pd.testing.assert_frame_equal(reverse, data)



import datetime as dt

import numpy as np
import pandas as pd

from rdt.performance.datasets import datetime


class TestRandomGapDatetimeGenerator:

    def test(self):
        output = datetime.RandomGapDatetimeGenerator.generate(10)
        assert len(output) == 10
        assert output.dtype == 'datetime64[us]'
        assert len(pd.unique(output)) > 1
        assert np.isnan(output).sum() == 0


class TestRandomGapSecondsDatetimeGenerator:

    def test(self):
        output = datetime.RandomGapSecondsDatetimeGenerator.generate(10)
        assert len(output) == 10
        assert output.dtype == 'datetime64[us]'
        assert len(pd.unique(output)) > 1
        assert np.isnan(output).sum() == 0


class TestRandomGapDatetimeNaNsGenerator:

    def test(self):
        output = datetime.RandomGapDatetimeNaNsGenerator.generate(10)
        assert len(output) == 10
        assert output.dtype == 'O'
        assert len(pd.unique(output)) > 1
        assert pd.isna(output).sum() > 0


class TestEqualGapHoursDatetimeGenerator:

    def test(self):
        output = datetime.EqualGapHoursDatetimeGenerator.generate(10)
        assert len(output) == 10
        assert output.dtype == 'datetime64[us]'
        assert all(x == dt.timedelta(hours=1) for x in np.diff(output))
        assert np.isnan(output).sum() == 0


class TestEqualGapDaysDatetimeGenerator:

    def test(self):
        output = datetime.EqualGapDaysDatetimeGenerator.generate(10)
        assert len(output) == 10
        assert output.dtype == 'datetime64[us]'
        assert all(x == dt.timedelta(1) for x in np.diff(output))
        assert np.isnan(output).sum() == 0


class TestEqualGapWeeksDatetimeGenerator:

    def test(self):
        output = datetime.EqualGapWeeksDatetimeGenerator.generate(10)
        assert len(output) == 10
        assert output.dtype == 'datetime64[us]'
        assert all(x == dt.timedelta(weeks=1) for x in np.diff(output))
        assert np.isnan(output).sum() == 0



import pytest

from rdt.transformers import (
    BinaryEncoder, get_transformer_class, get_transformer_instance, get_transformer_name)
from rdt.transformers.addons.identity.identity import IdentityTransformer


def test_get_transformer_name():
    """Test the ``get_transformer_name`` method.

    Validate the method returns the class path when passed the class.

    Input:
        - a class.

    Output:
        - the path of the class.
    """
    # Setup
    transformer = BinaryEncoder

    # Run
    returned = get_transformer_name(transformer)

    # Assert
    assert returned == 'rdt.transformers.boolean.BinaryEncoder'


def test_get_transformer_name_incorrect_input():
    """Test the ``get_transformer_name`` method crashes.

    Validate the method raises a ``ValueError`` when passed a string.

    Input:
        - a string.

    Raises:
        - ``ValueError``, with the correct output message.
    """
    # Setup
    transformer = 'rdt.transformers.boolean.BinaryEncoder'

    # Run / Assert
    error_msg = 'The transformer rdt.transformers.boolean.BinaryEncoder must be passed as a class.'
    with pytest.raises(ValueError, match=error_msg):
        get_transformer_name(transformer)


def test_get_transformer_class_transformer_path():
    """Test the ``get_transformer_class`` method.

    Validate the method returns the correct class when passed the class path.

    Input:
        - a string describing the transformer path.

    Output:
        - the class corresponding to the transformer path.
    """
    # Setup
    transformer_path = 'rdt.transformers.boolean.BinaryEncoder'

    # Run
    returned = get_transformer_class(transformer_path)

    # Assert
    assert returned == BinaryEncoder


def test_get_transformer_class_transformer_path_addon():
    """Test the ``get_transformer_class`` method.

    Validate the method returns the correct class when passed an addon path.

    Input:
        - a string describing the transformer path.

    Output:
        - the class corresponding to the transformer path.
    """
    # Setup
    transformer_path = 'rdt.transformers.addons.identity.identity.IdentityTransformer'

    # Run
    returned = get_transformer_class(transformer_path)

    # Assert
    assert returned == IdentityTransformer


def test_get_transformer_instance_instance():
    transformer = BinaryEncoder(missing_value_replacement=None)

    returned = get_transformer_instance(transformer)

    assert isinstance(returned, BinaryEncoder)
    assert returned.missing_value_replacement is None


def test_get_transformer_instance_str():
    transformer = 'rdt.transformers.BinaryEncoder'

    returned = get_transformer_instance(transformer)

    assert isinstance(returned, BinaryEncoder)


def test_get_transformer_instance_addon():
    transformer = 'rdt.transformers.addons.identity.identity.IdentityTransformer'

    returned = get_transformer_instance(transformer)

    assert isinstance(returned, IdentityTransformer)


def test_get_transformer_instance_class():
    transformer = BinaryEncoder

    returned = get_transformer_instance(transformer)

    assert isinstance(returned, BinaryEncoder)



import numpy as np
import pandas as pd

from rdt.transformers.datetime import OptimizedTimestampEncoder, UnixTimestampEncoder


class TestUnixTimestampEncoder:
    def setup_method(self):
        self.random_state = np.random.get_state()
        np.random.set_state(np.random.RandomState(7).get_state())

    def teardown_method(self):
        np.random.set_state(self.random_state)

    def test_unixtimestampencoder(self):
        ute = UnixTimestampEncoder(missing_value_replacement='mean')
        data = pd.DataFrame({'column': pd.to_datetime([None, '1996-10-17', '1965-05-23'])})

        # Run
        ute.fit(data, column='column')
        transformed = ute.transform(data)
        reverted = ute.reverse_transform(transformed)

        # Asserts
        expected_transformed = pd.DataFrame({
            'column.value': [3.500064e+17, 845510400000000000, -145497600000000000]
        })

        pd.testing.assert_frame_equal(expected_transformed, transformed)
        pd.testing.assert_frame_equal(reverted, data)

    def test_unixtimestampencoder_different_format(self):
        ute = UnixTimestampEncoder(missing_value_replacement='mean', datetime_format='%b %d, %Y')
        data = pd.DataFrame({'column': [None, 'Oct 17, 1996', 'May 23, 1965']})

        # Run
        ute.fit(data, column='column')
        transformed = ute.transform(data)
        reverted = ute.reverse_transform(transformed)

        # Asserts
        expect_transformed = pd.DataFrame({
            'column.value': [3.500064e+17, 845510400000000000, -145497600000000000]
        })
        pd.testing.assert_frame_equal(expect_transformed, transformed)
        pd.testing.assert_frame_equal(reverted, data)


class TestOptimizedTimestampEncoder:
    def setup_method(self):
        self.random_state = np.random.get_state()
        np.random.set_state(np.random.RandomState(7).get_state())

    def teardown_method(self):
        np.random.set_state(self.random_state)

    def test_optimizedtimestampencoder(self):
        ote = OptimizedTimestampEncoder(missing_value_replacement='mean')
        data = pd.DataFrame({'column': pd.to_datetime([None, '1996-10-17', '1965-05-23'])})

        # Run
        ote.fit(data, column='column')
        transformed = ote.transform(data)
        reverted = ote.reverse_transform(transformed)

        # Asserts
        expect_transformed = pd.DataFrame({'column.value': [4051.0, 9786.0, -1684.0]})
        pd.testing.assert_frame_equal(expect_transformed, transformed)
        pd.testing.assert_frame_equal(reverted, data)



import platform
from unittest.mock import Mock, patch

import numpy as np
import pandas as pd
import pytest

from rdt.transformers.datetime import OptimizedTimestampEncoder, UnixTimestampEncoder
from rdt.transformers.null import NullTransformer


class TestUnixTimestampEncoder:

    def test___init__(self):
        """Test the ``__init__`` method.

        Validate the passed arguments are stored as attributes.

        Setup:
            - initialize a ``UnixTimestampEncoder`` with values for each parameter.

        Side effect:
            - the ``missing_value_replacement`` attribute has been assigned as ``'mode'``.
            - the ``model_missing_values`` attribute has been assigned as True.
            - the ``datetime_format`` attribute has been assigned.
        """
        # Setup
        transformer = UnixTimestampEncoder(
            missing_value_replacement='mode',
            model_missing_values=True,
            datetime_format='%M-%d-%Y'
        )

        # Asserts
        assert transformer.missing_value_replacement == 'mode'
        assert transformer.model_missing_values is True
        assert transformer.datetime_format == '%M-%d-%Y'

    def test_is_composition_identity_null_transformer_true(self):
        """Test the ``is_composition_identity`` method with a ``null_transformer``.

        When the attribute ``null_transformer`` is not None and a null column is not created,
        this method should simply return False.

        Setup:
            - initialize a ``UnixTimestampEncoder`` transformer which sets
            ``self.null_transformer`` to a ``NullTransformer``.

        Output:
            - False.
        """
        # Setup
        transformer = UnixTimestampEncoder()
        transformer.null_transformer = NullTransformer(missing_value_replacement='fill')

        # Run
        output = transformer.is_composition_identity()

        # Assert
        assert output is False

    def test_is_composition_identity_null_transformer_false(self):
        """Test the ``is_composition_identity`` method without a ``null_transformer``.

        When the attribute ``null_transformer`` is None, this method should return
        the value stored in the ``COMPOSITION_IS_IDENTITY`` attribute.

        Setup:
            - initialize a ``UnixTimestampEncoder`` transformer which sets
            ``self.null_transformer`` to None.

        Output:
            - the value stored in ``self.COMPOSITION_IS_IDENTITY``.
        """
        # Setup
        transformer = UnixTimestampEncoder()
        transformer.null_transformer = None

        # Run
        output = transformer.is_composition_identity()

        # Assert
        assert output is True

    def test_get_output_sdtypes(self):
        """Test the ``get_output_sdtypes`` method when a null column is created.

        When a null column is created, this method should apply the ``_add_prefix``
        method to the following dictionary of output sdtypes:

        output_sdtypes = {
            'value': 'float',
            'is_null': 'float'
        }

        Setup:
            - initialize a ``UnixTimestampEncoder`` transformer which:
                - sets ``self.null_transformer`` to a ``NullTransformer`` where
                ``self._model_missing_values`` is True.
                - sets ``self.column_prefix`` to a column name.

        Output:
            - the ``output_sdtypes`` dictionary, but with the ``self.column_prefix``
            added to the beginning of the keys.
        """
        # Setup
        transformer = UnixTimestampEncoder()
        transformer.null_transformer = NullTransformer(missing_value_replacement='fill')
        transformer.null_transformer._model_missing_values = True
        transformer.column_prefix = 'a#b'

        # Run
        output = transformer.get_output_sdtypes()

        # Assert
        expected = {
            'a#b.value': 'float',
            'a#b.is_null': 'float'
        }
        assert output == expected

    def test__convert_to_datetime(self):
        """Test the ``_convert_to_datetime`` method.

        Test to make sure the transformer converts the data to datetime
        if it is of type ``object`` and can be converted.

        Input:
            - a pandas Series of dtype object, with elements that can be
            converted to datetime.

        Output:
            - a pandas series of type datetime.
        """
        # Setup
        data = pd.Series(['2020-01-01', '2020-02-01', '2020-03-01'])
        transformer = UnixTimestampEncoder()

        # Run
        converted_data = transformer._convert_to_datetime(data)

        # Assert
        expected_data = pd.Series(pd.to_datetime(['2020-01-01', '2020-02-01', '2020-03-01']))
        pd.testing.assert_series_equal(expected_data, converted_data)

    def test__convert_to_datetime_format(self):
        """Test the ``_convert_to_datetime`` method.

        Test to make sure the transformer uses the specified format
        to convert the data to datetime.

        Setup:
            - The transformer will be initialized with a format.
        Input:
            - a pandas Series of dtype object, with elements in the specified
            format.

        Output:
            - a pandas series of type datetime.
        """
        # Setup
        data = pd.Series(['01Feb2020', '02Mar2020', '03Jan2010'])
        dt_format = '%d%b%Y'
        transformer = UnixTimestampEncoder(datetime_format=dt_format)

        # Run
        converted_data = transformer._convert_to_datetime(data)

        # Assert
        expected_data = pd.Series(pd.to_datetime(['01Feb2020', '02Mar2020', '03Jan2010']))
        pd.testing.assert_series_equal(expected_data, converted_data)

    def test__convert_to_datetime_not_convertible_raises_error(self):
        """Test the ``_convert_to_datetime`` method.

        Test to make sure a ``TypeError`` is raised if the data is of type
        ``object`` but can't be converted.

        Input:
            - a pandas Series of dtype object, with elements that can't be
            converted to datetime.

        Expected behavior:
            - a ``TypeError`` is raised.
        """
        # Setup
        data = pd.Series(['2020-01-01-can', '2020-02-01-not', '2020-03-01-convert'])
        transformer = UnixTimestampEncoder()

        # Run
        error_message = 'Data must be of dtype datetime, or castable to datetime.'
        with pytest.raises(TypeError, match=error_message):
            transformer._convert_to_datetime(data)

    def test__convert_to_datetime_wrong_format_raises_error(self):
        """Test the ``_convert_to_datetime`` method.

        Test to make sure the transformer raises an error if the data does
        not match the specified format.

        Setup:
            - The transformer will be initialized with a format.
        Input:
            - a pandas Series of dtype object, with elements in the wrong
            format.

        Output:
            - a pandas series of type datetime.
        """
        # Setup
        data = pd.Series(['01-02-2020', '02-03-2020', '03J-01-2010'])
        dt_format = '%d%b%Y'
        transformer = UnixTimestampEncoder(datetime_format=dt_format)

        # Run
        error_message = 'Data does not match specified datetime format.'
        with pytest.raises(ValueError, match=error_message):
            transformer._convert_to_datetime(data)

    def test__transform_helper_calls_convert_to_datetime(self):
        """Test the ``_transform_helper`` method.

        Validate the helper transformer produces the correct value with ``strip_constant`` True.

        Input:
            - a pandas series of datetimes.

        Output:
            - a pandas series of the transformed datetimes.
        """
        # Setup
        data = pd.to_datetime(['2020-01-01', '2020-02-01', '2020-03-01'])
        transformer = UnixTimestampEncoder()
        transformer._convert_to_datetime = Mock()
        transformer._convert_to_datetime.return_value = data

        # Run
        transformer._transform_helper(data)

        # Assert
        transformer._convert_to_datetime.assert_called_once_with(data)

    def test__transform_helper(self):
        """Test the ``_transform_helper`` method.

        Validate the helper transformer produces the correct values.

        Input:
            - a pandas series of datetimes.

        Output:
            - a pandas series of the transformed datetimes.
        """
        # Setup
        data = pd.to_datetime(['2020-01-01', '2020-02-01', '2020-03-01'])
        transformer = UnixTimestampEncoder()

        # Run
        transformed = transformer._transform_helper(data)

        # Assert
        np.testing.assert_allclose(transformed, np.array([
            1.577837e+18, 1.580515e+18, 1.583021e+18,
        ]), rtol=1e-5)

    def test__reverse_transform_helper_nulls(self):
        """Test the ``_reverse_transform_helper`` with null values.

        Setup:
            - Mock the ``instance.null_transformer``.
            - Set the ``missing_value_replacement``.

        Input:
            - a pandas series.

        Output:
            - a pandas datetime index.

        Expected behavior:
            - The mock should call its ``reverse_transform`` method.
        """
        # Setup
        data = pd.to_datetime(['2020-01-01', '2020-02-01', '2020-03-01'])
        transformer = UnixTimestampEncoder(missing_value_replacement='mean')
        transformer.null_transformer = Mock()
        transformer.null_transformer.reverse_transform.return_value = pd.Series([1, 2, 3])

        # Run
        transformer._reverse_transform_helper(data)

        # Assert
        transformer.null_transformer.reverse_transform.assert_called_once()
        datetimes = transformer.null_transformer.reverse_transform.mock_calls[0][1][0]
        np.testing.assert_array_equal(data.to_numpy(), datetimes)

    @patch('rdt.transformers.datetime.NullTransformer')
    def test__fit(self, null_transformer_mock):
        """Test the ``_fit`` method for numpy arrays.

        Validate that this method (1) sets ``self.null_transformer`` to the
        ``NullTransformer`` with the correct attributes, (2) calls the
        ``self.null_transformer``'s ``fit`` method, and (3) calls the
        ``_transform_helper`` method with the transformed data.

        Input:
            - a pandas Series.

        Side effects:
            - sets ``self.null_transformer`` to the ``NullTransformer`` with
            the correct attributes.
            - calls the ``self.null_transformer``'s ``fit`` method.
            - calls the ``_transform_helper`` method with the transformed data.
        """
        # Setup
        data = pd.to_datetime(['2020-01-01', '2020-02-01', '2020-03-01'])
        transformer = UnixTimestampEncoder()

        # Run
        transformer._fit(data)

        # Assert
        null_transformer_mock.assert_called_once_with(None, False)
        assert null_transformer_mock.return_value.fit.call_count == 1
        np.testing.assert_allclose(
            null_transformer_mock.return_value.fit.call_args_list[0][0][0],
            np.array([1.577837e+18, 1.580515e+18, 1.583021e+18]), rtol=1e-5
        )

    def test__fit_calls_transform_helper(self):
        """Test the ``_fit`` method.

        The ``_fit`` method should call the ``_transform_helper`` method.
        """
        # Setup
        data = pd.to_datetime(['2020-01-01', '2020-02-01', '2020-03-01'])
        transformer = UnixTimestampEncoder()
        transformer._transform_helper = Mock()
        transformer._transform_helper.return_value = pd.Series(data)

        # Run
        transformer._fit(data)

        # Assert
        transformer._transform_helper.assert_called_once()

    @patch('rdt.transformers.datetime._guess_datetime_format_for_array')
    def test__fit_calls_guess_datetime_format(self, mock__guess_datetime_format_for_array):
        """Test the ``_fit`` method.

        The ``_fit`` method should call the ``_transform_helper`` method.
        """
        # Setup
        data = pd.Series(['2020-02-01', '2020-03-01'])
        mock__guess_datetime_format_for_array.return_value = '%Y-%m-%d'
        transformer = UnixTimestampEncoder()

        # Run
        transformer._fit(data)

        # Assert
        np.testing.assert_array_equal(
            mock__guess_datetime_format_for_array.call_args[0][0],
            np.array(['2020-02-01', '2020-03-01'])
        )
        assert transformer.datetime_format == '%Y-%m-%d'

    def test__transform(self):
        """Test the ``_transform`` method for numpy arrays.

        Validate that this method calls the ``null_transformer.transform`` method.
        It should also check that the final output is correct.

        Setup:
            - mock behavior of ``_fit``.

        Input:
            - a pandas Series.

        Output:
            - a numpy array containing the transformed data.

        Side effect:
            - calls the ``null_transformer.transform`` method with the transformed data.
        """
        # Setup
        data = pd.to_datetime(['2020-01-01', '2020-02-01', '2020-03-01'])
        transformer = UnixTimestampEncoder()
        transformer.null_transformer = Mock()

        # Run
        transformer._transform(data)

        # Assert
        assert transformer.null_transformer.transform.call_count == 1
        np.testing.assert_allclose(
            transformer.null_transformer.transform.call_args_list[0][0],
            np.array([[1.577837e+18, 1.580515e+18, 1.583021e+18]]), rtol=1e-5
        )

    def test__reverse_transform_all_none(self):
        """Test the ``_reverse_transform`` method with ``None`` values.

        Validate that the method transforms ``None`` into ``NaT``.

        Input:
            - A ``pd.Series`` with ``None`` as a value.

        Output:
            - A ``DatetimeIndex`` with ``NaT`` as a value.
        """
        # Setup
        ute = UnixTimestampEncoder()

        # Run
        output = ute._reverse_transform(pd.Series([None]))

        # Assert
        expected = pd.Series(pd.to_datetime(['NaT']))
        pd.testing.assert_series_equal(output, expected)

    def test__reverse_transform(self):
        """Test the ``_reverse_transform`` method.

        Validate that the method correctly reverse transforms.

        Input:
            - a numpy array.

        Output:
            - a pandas ``DatetimeIndex`` of the correct datetimes.
        """
        # Setup
        ute = UnixTimestampEncoder(missing_value_replacement=None)
        transformed = np.array([1.5778368e+18, 1.5805152e+18, 1.5830208e+18])

        # Run
        output = ute._reverse_transform(transformed)

        # Assert
        expected = pd.Series(pd.to_datetime(['2020-01-01', '2020-02-01', '2020-03-01']))
        pd.testing.assert_series_equal(output, expected)

    def test__reverse_transform_datetime_format(self):
        """Test the ``_reverse_transform`` method returns the correct datetime format.

        Setup:
            - Set the instance to have a different ``datetime_format``.

        Input:
            - a numpy array of integers.

        Output:
            - a pandas ``Series`` of the datetimes in the right format.
        """
        # Setup
        ute = UnixTimestampEncoder(missing_value_replacement=None)
        ute.datetime_format = '%b %d, %Y'
        transformed = np.array([1.5778368e+18, 1.5805152e+18, 1.5830208e+18])

        # Run
        output = ute._reverse_transform(transformed)

        # Assert
        expected = pd.Series(['Jan 01, 2020', 'Feb 01, 2020', 'Mar 01, 2020'])
        pd.testing.assert_series_equal(output, expected)

    def test__reverse_transform_datetime_format_with_strftime_formats(self):
        """Test the ``_reverse_transform`` method returns the correct datetime format.

        Setup:
            - Set the instance to have a different ``datetime_format`` which includes `-` on it.

        Input:
            - a numpy array of integers.

        Output:
            - a pandas ``Series`` of the datetimes in the right format.
        """
        # Setup
        ute = UnixTimestampEncoder(missing_value_replacement=None)
        ute.datetime_format = '%b %-d, %Y'
        transformed = np.array([1.5778368e+18, 1.5805152e+18, 1.5830208e+18])

        # Run
        output = ute._reverse_transform(transformed)

        # Assert
        expected = pd.Series(['Jan 1, 2020', 'Feb 1, 2020', 'Mar 1, 2020'])
        if 'windows' not in platform.system().lower():
            pd.testing.assert_series_equal(output, expected)


class TestOptimizedTimestampEncoder:

    def test__find_divider(self):
        """Test the ``_find_divider`` method.

        Find the greatest common denominator out of these values: [10] * 9 + [60, 60, 24],
        where each consecutive value in the list is multiplied by the previous one
        (so 10, 100, 1000, etc).

        Input:
            - a numpy array.

        Side effect:
            - sets ``self.divider`` to the correct divider.
        """
        # Setup
        data = np.array([100, 7919])
        transformer = OptimizedTimestampEncoder()

        # Run
        transformer._find_divider(data)

        # Assert
        assert transformer.divider == 1

    def test__transform_helper(self):
        """Test the ``_transform_helper`` method.

        Validate the helper method produces the values stripped to the smallest
        non-zero time unit.

        Input:
            - a pandas series of datetimes.

        Output:
            - a pandas series of the transformed datetimes.
        """
        # Setup
        data = pd.to_datetime(['2020-01-01', '2020-02-01', '2020-03-01'])
        transformer = OptimizedTimestampEncoder()

        # Run
        transformed = transformer._transform_helper(data)

        # Assert
        np.testing.assert_allclose(transformed, np.array([
            18262., 18293., 18322.,
        ]))

    def test__reverse_transform_helper(self):
        """Test the ``_reverse_transform_helper`` method.

        Validate the helper produces the values multiplied by the
        smallest non-zero time unit.

        Input:
            - a pandas series of timestamps.

        Output:
            - a numpy array of the values multiplied by ``self.divider``.
        """
        # Setup
        data = pd.Series([18262., 18293., 18322.])
        transformer = OptimizedTimestampEncoder()
        transformer.divider = 1000
        transformer.null_transformer = Mock()
        transformer.null_transformer.reverse_transform.side_effect = lambda x: x

        # Run
        multiplied = transformer._reverse_transform_helper(data)

        # Assert
        expected = np.array([18262000, 18293000, 18322000])
        np.testing.assert_allclose(multiplied, expected)
